import os
import shutil
from glob import glob

import packaging.version
from qtpy.QtWidgets import QMessageBox, QWidget

from PartSeg import parsed_version, state_store
from PartSeg._launcher.check_version import IGNORE_FILE
from PartSeg.common_backend import napari_get_settings


def _parse_version(name):
    try:
        return packaging.version.parse(name)
    except packaging.version.InvalidVersion:
        return None


def import_config():
    """
    Check if settings folder for current version already exists.

    Otherwise, when settings for previous PartSeg version exists,
    ask user if he would to import settings from the newest one.
    """
    if os.path.exists(state_store.save_folder):
        return
    version = packaging.version.parse(parsed_version.base_version)
    base_folder = os.path.dirname(state_store.save_folder)
    possible_folders = glob(os.path.join(base_folder, "*"))
    versions = sorted(
        (
            x
            for x in [_parse_version(os.path.basename(y)) for y in possible_folders]
            if isinstance(x, packaging.version.Version)
        ),
        reverse=True,
    )
    before_version = next((x for x in versions if x < version), None)
    if before_version is not None:
        before_name = str(before_version)
        widget = QWidget()
        resp = QMessageBox.question(
            widget,
            "Import from old version",
            "There is no configuration folder for this version of PartSeg\n"
            "Would you like to import it from " + before_name + " version of PartSeg",
            QMessageBox.Yes | QMessageBox.No,
            QMessageBox.Yes,
        )
        if resp == QMessageBox.Yes:
            shutil.copytree(os.path.join(base_folder, before_name), state_store.save_folder)
            if os.path.exists(os.path.join(state_store.save_folder, IGNORE_FILE)):
                os.remove(os.path.join(state_store.save_folder, IGNORE_FILE))
            napari_settings = napari_get_settings(state_store.save_folder)
            if hasattr(napari_settings, "load") and napari_settings.load is not None:
                napari_settings.load()
            elif getattr(napari_settings, "_load", None) is not None:
                napari_settings._load()  # pylint: disable=protected-access



# pylint: disable=no-self-use
from copy import copy
from unittest.mock import patch

import pytest
from qtpy.QtWidgets import QTreeWidgetItem

from PartSeg._roi_analysis import prepare_plan_widget
from PartSegCore import Units
from PartSegCore.algorithm_describe_base import ROIExtractionProfile
from PartSegCore.analysis import AnalysisAlgorithmSelection, SegmentationPipeline, SegmentationPipelineElement
from PartSegCore.analysis.calculation_plan import CalculationTree, MeasurementCalculate, NodeType, RootType, Save
from PartSegCore.analysis.save_functions import SaveAsTiff
from PartSegCore.io_utils import SaveMaskAsTiff
from PartSegCore.mask_create import MaskProperty
from PartSegCore.segmentation.restartable_segmentation_algorithms import (
    LowerThresholdAlgorithm,
    UpperThresholdAlgorithm,
)

NOT_FOUND_STR = "not found in register"


@pytest.mark.parametrize(
    ("mask1", "mask2", "enabled"),
    [
        ("mask1", "mask2", True),
        ("mask1", "", False),
        ("", "mask2", False),
        ("", "", False),
        ("mask1", "mask1", False),
        ("mask2", "mask2", False),
        ("mask2", "mask1", True),
        ("mask", "mask2", False),
        ("mask2", "mask", False),
        ("mask", "mask1", False),
        ("mask1", "mask", False),
    ],
)
def test_two_mask_dialog(qtbot, mask1, mask2, enabled):
    dialog = prepare_plan_widget.TwoMaskDialog(["mask1", "mask2"])
    qtbot.addWidget(dialog)
    assert not dialog.ok_btn.isEnabled()
    dialog.mask1_name.setText(mask1)
    dialog.mask2_name.setText(mask2)
    assert dialog.ok_btn.isEnabled() is enabled
    assert dialog.get_result() == (mask1, mask2)


def test_two_mask_dialog_strip(qtbot):
    dialog = prepare_plan_widget.TwoMaskDialog(["mask1", "mask2"])
    qtbot.addWidget(dialog)
    dialog.mask1_name.setText("mask1 ")
    dialog.mask2_name.setText(" mask2")
    assert dialog.get_result() == ("mask1", "mask2")


@pytest.mark.parametrize(
    ("mask", "enabled"),
    [
        ("mask", False),
        ("", False),
        ("mask1", True),
        ("mask2", True),
    ],
)
def test_mask_dialog(qtbot, mask, enabled):
    dialog = prepare_plan_widget.MaskDialog(["mask1", "mask2"])
    qtbot.addWidget(dialog)
    assert not dialog.ok_btn.isEnabled()
    dialog.mask1_name.setText(mask)
    assert dialog.ok_btn.isEnabled() is enabled
    assert dialog.get_result() == (mask,)


def test_mask_dialog_strip(qtbot):
    dialog = prepare_plan_widget.MaskDialog(["mask1", "mask2"])
    qtbot.addWidget(dialog)
    dialog.mask1_name.setText("mask1 ")
    assert dialog.get_result() == ("mask1",)


class TestFileMaskWidget:
    def test_create(self, qtbot):
        widget = prepare_plan_widget.FileMask()
        qtbot.addWidget(widget)
        assert widget.select_type.currentText() == "Suffix"

    @pytest.mark.parametrize("name", ["", "mask"])
    def test_suffix_state(self, qtbot, name):
        widget = prepare_plan_widget.FileMask()
        qtbot.addWidget(widget)
        res = widget.get_value(name=name)
        assert isinstance(res, prepare_plan_widget.MaskSuffix)
        assert res.name == name
        assert res.suffix == "_mask"
        assert widget.is_valid()

    @pytest.mark.parametrize("name", ["", "mask"])
    def test_replace_path_state(self, qtbot, name):
        widget = prepare_plan_widget.FileMask()
        widget.show()
        qtbot.addWidget(widget)
        assert not widget.second_text.isVisible()
        with qtbot.waitSignal(widget.select_type.currentIndexChanged):
            widget.select_type.setCurrentIndex(1)
        widget.first_text.setText("val")
        widget.second_text.setText("val2")
        assert widget.second_text.isVisible()
        res = widget.get_value(name=name)
        assert isinstance(res, prepare_plan_widget.MaskSub)
        assert res.name == name
        assert res.base == "val"
        assert res.rep == "val2"
        assert widget.is_valid()
        widget.hide()

    @pytest.mark.parametrize("name", ["", "mask"])
    def test_mapping_state(self, qtbot, name, tmp_path):
        file_path = str(tmp_path / "file.txt")
        with open(file_path, "w") as f:
            f.write("test")

        widget = prepare_plan_widget.FileMask()
        widget.show()
        qtbot.addWidget(widget)
        with qtbot.waitSignal(widget.select_type.currentIndexChanged):
            widget.select_type.setCurrentIndex(2)
        widget.first_text.setText(file_path)
        res = widget.get_value(name=name)
        assert isinstance(res, prepare_plan_widget.MaskFile)
        assert res.name == name
        assert res.path_to_file == file_path
        assert widget.is_valid()

    def test_change_state_memory(self, qtbot):
        widget = prepare_plan_widget.FileMask()
        qtbot.addWidget(widget)
        widget.change_type(prepare_plan_widget.FileMaskType.Suffix)
        widget.first_text.setText("suffix_text")
        widget.change_type(prepare_plan_widget.FileMaskType.Replace)
        widget.first_text.setText("replace_base")
        widget.second_text.setText("replace")
        widget.change_type(prepare_plan_widget.FileMaskType.Mapping_file)
        widget.first_text.setText("file_path")
        widget.change_type(prepare_plan_widget.FileMaskType.Suffix)
        assert widget.first_text.text() == "suffix_text"
        widget.change_type(prepare_plan_widget.FileMaskType.Replace)
        assert widget.first_text.text() == "replace_base"
        assert widget.second_text.text() == "replace"
        widget.change_type(prepare_plan_widget.FileMaskType.Mapping_file)
        assert widget.first_text.text() == "file_path"

    def test_select_file(self, qtbot, monkeypatch):
        widget = prepare_plan_widget.FileMask()
        qtbot.addWidget(widget)
        with qtbot.waitSignal(widget.select_type.currentIndexChanged):
            widget.select_type.setCurrentIndex(2)
        monkeypatch.setattr(prepare_plan_widget.QFileDialog, "exec_", lambda x: True)
        monkeypatch.setattr(prepare_plan_widget.QFileDialog, "selectedFiles", lambda x: ["file_path"])
        assert widget.first_text.text() == ""
        widget.select_file()
        assert widget.first_text.text() == "file_path"


@pytest.fixture()
def calculation_plan(measurement_profiles):
    roi_extraction = AnalysisAlgorithmSelection.get_default()
    return prepare_plan_widget.CalculationPlan(
        tree=CalculationTree(
            operation=RootType.Image,
            children=[
                CalculationTree(
                    operation=ROIExtractionProfile(
                        name="test",
                        algorithm=roi_extraction.name,
                        values=roi_extraction.values,
                    ),
                    children=[
                        CalculationTree(
                            operation=MeasurementCalculate(
                                channel=0, units=Units.nm, name_prefix="", measurement_profile=measurement_profiles[0]
                            ),
                            children=[],
                        )
                    ],
                ),
            ],
        ),
        name="test",
    )


class TestCalculateInfo:
    def test_create(self, qtbot, part_settings, calculation_plan):
        part_settings.batch_plans["test"] = calculation_plan
        widget = prepare_plan_widget.CalculateInfo(part_settings)
        qtbot.addWidget(widget)
        assert widget.calculate_plans.count() == 1

    def test_select_plan(self, qtbot, part_settings, calculation_plan):
        part_settings.batch_plans["test"] = calculation_plan
        widget = prepare_plan_widget.CalculateInfo(part_settings)
        qtbot.addWidget(widget)
        assert widget.plan_view.topLevelItemCount() == 0
        with qtbot.waitSignal(widget.calculate_plans.currentRowChanged):
            widget.calculate_plans.setCurrentRow(0)
        assert widget.plan_view.topLevelItemCount() == 1

    def test_add_calculation_plan(self, qtbot, part_settings, calculation_plan):
        part_settings.batch_plans["test"] = calculation_plan
        widget = prepare_plan_widget.CalculateInfo(part_settings)
        qtbot.addWidget(widget)
        assert widget.calculate_plans.count() == 1
        widget.calculate_plans.setCurrentRow(0)
        calculation_plan2 = prepare_plan_widget.CalculationPlan(
            tree=calculation_plan.execution_tree,
            name="test2",
        )
        part_settings.batch_plans["test2"] = calculation_plan2
        assert widget.calculate_plans.count() == 2

    def test_add_bad_plan(self, qtbot, part_settings, calculation_plan):
        bad_plan = copy(calculation_plan)
        bad_plan.name = "test2"
        bad_plan.execution_tree.children[0].operation = {"__error__": NOT_FOUND_STR}
        part_settings.batch_plans[calculation_plan.name] = calculation_plan
        part_settings.batch_plans[bad_plan.name] = bad_plan

        widget = prepare_plan_widget.CalculateInfo(part_settings)
        qtbot.addWidget(widget)

        assert widget.calculate_plans.count() == 2
        assert widget.calculate_plans.item(0).text() == calculation_plan.name
        assert widget.calculate_plans.item(0).icon().isNull()
        assert widget.calculate_plans.item(1).text() == bad_plan.name
        assert not widget.calculate_plans.item(1).icon().isNull()

    def test_delete_plan(self, qtbot, part_settings, calculation_plan):
        part_settings.batch_plans["test"] = calculation_plan
        widget = prepare_plan_widget.CalculateInfo(part_settings)
        qtbot.addWidget(widget)
        assert widget.calculate_plans.count() == 1
        widget.delete_plan()
        assert widget.calculate_plans.count() == 1
        widget.calculate_plans.setCurrentRow(0)
        widget.delete_plan()
        assert widget.calculate_plans.count() == 0
        assert "test" not in part_settings.batch_plans

    def test_edit_plan(self, qtbot, part_settings, calculation_plan):
        part_settings.batch_plans["test"] = calculation_plan
        widget = prepare_plan_widget.CalculateInfo(part_settings)
        qtbot.addWidget(widget)
        with qtbot.assert_not_emitted(widget.plan_to_edit_signal):
            widget.edit_plan()
        widget.calculate_plans.setCurrentRow(0)
        with qtbot.waitSignal(widget.plan_to_edit_signal):
            widget.edit_plan()
        assert widget.plan_to_edit.name == "test"


class TestOtherOperations:
    def test_create(self, qtbot):
        widget = prepare_plan_widget.OtherOperations()
        qtbot.addWidget(widget)

    def test_change_root_type(self, qtbot):
        widget = prepare_plan_widget.OtherOperations()
        qtbot.addWidget(widget)
        with qtbot.waitSignal(widget.root_type_changed, check_params_cb=lambda x: x == RootType.Project):
            widget.change_root.setCurrentEnum(RootType.Project)

    def test_update_save_combo(self, qtbot):
        widget = prepare_plan_widget.OtherOperations()
        qtbot.addWidget(widget)
        with qtbot.waitSignal(widget.choose_save_method.currentTextChanged):
            widget.choose_save_method.setCurrentIndex(1)
        assert widget.save_btn.text().startswith("Save to")

        widget.save_changed("eeee")
        assert widget.save_btn.text() == "Save"
        assert widget.choose_save_method.currentIndex() == 0

    @pytest.mark.parametrize(("root_type", "replace"), [(NodeType.root, False), (NodeType.mask, True)])
    def test_set_current_node(self, qtbot, root_type, replace):
        widget = prepare_plan_widget.OtherOperations()
        qtbot.addWidget(widget)
        widget.set_current_node(None)
        assert not widget.save_btn.isEnabled()
        widget.set_replace(replace)
        widget.set_current_node(root_type, NodeType.root)
        assert not widget.save_btn.isEnabled()
        with qtbot.waitSignal(widget.choose_save_method.currentTextChanged):
            widget.choose_save_method.setCurrentText(SaveAsTiff.get_short_name())
        assert widget.save_btn.isEnabled()
        with qtbot.waitSignal(widget.choose_save_method.currentTextChanged):
            widget.choose_save_method.setCurrentText(SaveMaskAsTiff.get_short_name())
        assert not widget.save_btn.isEnabled()

    def test_save_action(self, qtbot, monkeypatch):
        monkeypatch.setattr(prepare_plan_widget.FormDialog, "exec_", lambda x: True)
        monkeypatch.setattr(
            prepare_plan_widget.FormDialog,
            "get_values",
            lambda x: {"suffix": "test", "directory": "test2", "sample": 1},
        )

        def check_save_params(params):
            assert params.suffix == "test"
            assert params.directory == "test2"
            assert params.algorithm == SaveAsTiff.get_name()
            assert params.short_name == SaveAsTiff.get_short_name()
            assert params.values == {"sample": 1}
            return True

        widget = prepare_plan_widget.OtherOperations()
        qtbot.addWidget(widget)
        widget.set_current_node(NodeType.root)
        with qtbot.waitSignal(widget.choose_save_method.currentTextChanged):
            widget.choose_save_method.setCurrentText(SaveAsTiff.get_short_name())
        with qtbot.waitSignal(widget.save_operation, check_params_cb=check_save_params):
            widget.save_btn.click()


class TestROIExtractionOp:
    def test_create(self, qtbot, part_settings):
        widget = prepare_plan_widget.ROIExtractionOp(settings=part_settings)
        qtbot.addWidget(widget)

    def test_selected_profile(self, qtbot, part_settings):
        def check_profile(name):
            def _check_profile(profile: ROIExtractionProfile):
                assert profile == part_settings.roi_profiles[name]
                return True

            return _check_profile

        part_settings.roi_profiles["test"] = ROIExtractionProfile(
            name="test",
            algorithm=LowerThresholdAlgorithm.get_name(),
            values=LowerThresholdAlgorithm.get_default_values(),
        )
        part_settings.roi_profiles["test2"] = ROIExtractionProfile(
            name="test2",
            algorithm=UpperThresholdAlgorithm.get_name(),
            values=UpperThresholdAlgorithm.get_default_values(),
        )
        widget = prepare_plan_widget.ROIExtractionOp(settings=part_settings)
        qtbot.addWidget(widget)

        assert widget.roi_profile.count() == 2
        with qtbot.assert_not_emitted(widget.roi_extraction_profile_add):
            widget._add_profile()
        widget.roi_extraction_tab.setCurrentWidget(widget.roi_profile)
        with qtbot.waitSignal(widget.roi_extraction_profile_selected, check_params_cb=check_profile("test")):
            widget.roi_profile.setCurrentRow(0)

        with qtbot.waitSignal(widget.roi_extraction_profile_selected, check_params_cb=check_profile("test2")):
            widget.roi_profile.setCurrentRow(1)

        with widget.enable_protect(), qtbot.assert_not_emitted(widget.roi_extraction_profile_selected):
            widget.roi_profile.setCurrentRow(0)

        part_settings.roi_profiles["test3"] = ROIExtractionProfile(
            name="test3",
            algorithm=LowerThresholdAlgorithm.get_name(),
            values=LowerThresholdAlgorithm.get_default_values(),
        )
        assert widget.roi_profile.count() == 3

        with qtbot.waitSignal(widget.roi_extraction_profile_add, check_params_cb=check_profile("test")):
            widget._add_profile()

    def test_select_pipeline(self, qtbot, part_settings):
        def check_pipeline(name):
            def _check_pipeline(pipeline: SegmentationPipeline):
                assert pipeline == part_settings.roi_pipelines[name]
                return True

            return _check_pipeline

        part_settings.roi_pipelines["test"] = SegmentationPipeline(
            name="test",
            segmentation=ROIExtractionProfile(
                name="test3",
                algorithm=LowerThresholdAlgorithm.get_name(),
                values=LowerThresholdAlgorithm.get_default_values(),
            ),
            mask_history=[
                SegmentationPipelineElement(
                    segmentation=ROIExtractionProfile(
                        name="test4",
                        algorithm=LowerThresholdAlgorithm.get_name(),
                        values=LowerThresholdAlgorithm.get_default_values(),
                    ),
                    mask_property=MaskProperty.simple_mask(),
                )
            ],
        )
        part_settings.roi_pipelines["test2"] = SegmentationPipeline(
            name="test2",
            segmentation=ROIExtractionProfile(
                name="test5",
                algorithm=LowerThresholdAlgorithm.get_name(),
                values=LowerThresholdAlgorithm.get_default_values(),
            ),
            mask_history=[
                SegmentationPipelineElement(
                    segmentation=ROIExtractionProfile(
                        name="test4",
                        algorithm=UpperThresholdAlgorithm.get_name(),
                        values=UpperThresholdAlgorithm.get_default_values(),
                    ),
                    mask_property=MaskProperty.simple_mask(),
                )
            ],
        )

        widget = prepare_plan_widget.ROIExtractionOp(part_settings)
        qtbot.addWidget(widget)
        with qtbot.waitSignal(widget.roi_extraction_tab.currentChanged):
            widget.roi_extraction_tab.setCurrentWidget(widget.roi_pipeline)
        assert widget.roi_pipeline.count() == 2
        with qtbot.assert_not_emitted(widget.roi_extraction_pipeline_add):
            widget._add_profile()
        with qtbot.waitSignal(widget.roi_extraction_pipeline_selected, check_params_cb=check_pipeline("test")):
            widget.roi_pipeline.setCurrentRow(0)
        with qtbot.waitSignal(widget.roi_extraction_pipeline_selected, check_params_cb=check_pipeline("test2")):
            widget.roi_pipeline.setCurrentRow(1)

        with widget.enable_protect(), qtbot.assert_not_emitted(widget.roi_extraction_pipeline_selected):
            widget.roi_pipeline.setCurrentRow(0)
        with qtbot.waitSignal(widget.roi_extraction_pipeline_add, check_params_cb=check_pipeline("test")):
            widget._add_profile()

    def test_set_node_type(self, qtbot, part_settings):
        part_settings.roi_profiles["test"] = ROIExtractionProfile(
            name="test",
            algorithm=LowerThresholdAlgorithm.get_name(),
            values=LowerThresholdAlgorithm.get_default_values(),
        )
        widget = prepare_plan_widget.ROIExtractionOp(part_settings)
        qtbot.addWidget(widget)
        assert not widget.choose_profile_btn.isEnabled()
        widget.set_current_node(NodeType.root)
        assert not widget.choose_profile_btn.isEnabled()
        widget.set_current_node(None)
        with qtbot.waitSignal(widget.roi_extraction_profile_selected):
            widget.roi_profile.setCurrentRow(0)
        assert not widget.choose_profile_btn.isEnabled()
        widget.set_current_node(NodeType.root)
        assert widget.choose_profile_btn.isEnabled()
        widget.set_current_node(NodeType.mask)
        assert widget.choose_profile_btn.isEnabled()
        widget.set_current_node(NodeType.file_mask)
        assert widget.choose_profile_btn.isEnabled()
        with qtbot.waitSignal(widget.roi_extraction_tab.currentChanged):
            widget.roi_extraction_tab.setCurrentWidget(widget.roi_pipeline)
        assert not widget.choose_profile_btn.isEnabled()

    def test_replace(self, qtbot, part_settings):
        widget = prepare_plan_widget.ROIExtractionOp(part_settings)
        qtbot.addWidget(widget)
        assert widget.choose_profile_btn.text() == "Add Profile"
        widget.set_replace(True)
        assert widget.choose_profile_btn.text() == "Replace Profile"
        widget.set_replace(False)
        assert widget.choose_profile_btn.text() == "Add Profile"
        with qtbot.waitSignal(widget.roi_extraction_tab.currentChanged):
            widget.roi_extraction_tab.setCurrentWidget(widget.roi_pipeline)
        assert widget.choose_profile_btn.text() == "Add Pipeline"
        widget.set_replace(True)
        assert widget.choose_profile_btn.text() == "Replace Pipeline"


class TestSelectMeasurementOp:
    def test_create(self, qtbot, part_settings):
        widget = prepare_plan_widget.SelectMeasurementOp(part_settings)
        qtbot.addWidget(widget)

    def test_replace(self, qtbot, part_settings):
        widget = prepare_plan_widget.SelectMeasurementOp(part_settings)
        qtbot.addWidget(widget)
        widget.set_replace(True)
        assert widget.add_measurement_btn.text() == "Replace set of measurements"
        widget.set_replace(False)
        assert widget.add_measurement_btn.text() == "Add set of measurements"

    def test_selected_profile(self, qtbot, part_settings):
        def check_measurement(name):
            def _check_measurement(measurement):
                return measurement == part_settings.measurement_profiles[name]

            return _check_measurement

        widget = prepare_plan_widget.SelectMeasurementOp(part_settings)
        qtbot.addWidget(widget)
        with qtbot.waitSignal(widget.set_of_measurement_selected, check_params_cb=check_measurement("statistic1")):
            widget.measurements_list.setCurrentRow(0)
        with qtbot.waitSignal(widget.set_of_measurement_selected, check_params_cb=check_measurement("statistic2")):
            widget.measurements_list.setCurrentRow(1)
        with widget.enable_protect(), qtbot.assert_not_emitted(widget.set_of_measurement_selected):
            widget.measurements_list.setCurrentRow(0)

    def test_set_current_node(self, qtbot, part_settings):
        widget = prepare_plan_widget.SelectMeasurementOp(part_settings)
        qtbot.addWidget(widget)
        widget.set_current_node(NodeType.segment)
        assert not widget.add_measurement_btn.isEnabled()
        widget.measurements_list.setCurrentRow(0)
        widget.set_current_node(NodeType.root)
        assert not widget.add_measurement_btn.isEnabled()
        widget.set_current_node(None)
        assert not widget.add_measurement_btn.isEnabled()
        widget.set_current_node(NodeType.segment)
        assert widget.add_measurement_btn.isEnabled()
        widget.set_current_node(NodeType.mask)
        assert not widget.add_measurement_btn.isEnabled()
        widget.set_current_node(NodeType.measurement)
        assert not widget.add_measurement_btn.isEnabled()
        widget.set_replace(True)
        assert widget.add_measurement_btn.isEnabled()
        widget.set_current_node(NodeType.segment)
        assert not widget.add_measurement_btn.isEnabled()

    def test_add_measurement(self, qtbot, part_settings):
        def check_measurement(measurement: prepare_plan_widget.MeasurementCalculate):
            mes = copy(part_settings.measurement_profiles["statistic1"])
            mes.name_prefix = "prefix_"
            assert measurement.measurement_profile == mes
            assert measurement.channel == 4
            assert measurement.name_prefix == "prefix_"
            return True

        widget = prepare_plan_widget.SelectMeasurementOp(part_settings)
        qtbot.addWidget(widget)
        assert widget.measurements_list.count() == 2
        with qtbot.assert_not_emitted(widget.set_of_measurement_add):
            widget._measurement_add()
        widget.measurements_list.setCurrentRow(0)
        widget.measurement_name_prefix.setText("prefix_")
        widget.choose_channel_for_measurements.setCurrentIndex(5)
        with qtbot.waitSignal(widget.set_of_measurement_add, check_params_cb=check_measurement):
            widget._measurement_add()


class TestSelectMaskOp:
    def test_create(self, qtbot, part_settings):
        widget = prepare_plan_widget.SelectMaskOp(part_settings)
        qtbot.addWidget(widget)

    def test_replace(self, qtbot, part_settings):
        widget = prepare_plan_widget.SelectMaskOp(part_settings)
        qtbot.addWidget(widget)
        widget.set_replace(True)
        assert widget.add_mask_btn.text().startswith("Replace")
        widget.set_replace(False)
        assert widget.add_mask_btn.text().startswith("Add")

    @pytest.mark.parametrize("node_type", NodeType.__members__.values())
    def test_set_current_node(self, qtbot, part_settings, node_type):
        widget = prepare_plan_widget.SelectMaskOp(part_settings)
        qtbot.addWidget(widget)
        widget.set_current_node(node_type)
        widget.mask_tab_select.setCurrentWidget(widget.file_mask)
        assert widget.add_mask_btn.isEnabled() == (node_type is NodeType.root)
        widget.mask_tab_select.setCurrentWidget(widget.mask_from_segmentation)
        assert widget.add_mask_btn.isEnabled() == (node_type is NodeType.segment)
        widget.mask_tab_select.setCurrentWidget(widget.mask_operation)
        assert widget.add_mask_btn.isEnabled() == (node_type is NodeType.root)

    @pytest.mark.parametrize("node_type", NodeType.__members__.values())
    def test_set_current_node_replace(self, qtbot, part_settings, node_type):
        widget = prepare_plan_widget.SelectMaskOp(part_settings)
        qtbot.addWidget(widget)
        widget.set_current_node(NodeType.mask, node_type)
        widget.mask_tab_select.setCurrentWidget(widget.file_mask)
        widget.set_replace(True)
        assert widget.add_mask_btn.isEnabled() == (node_type is NodeType.root)
        widget.mask_tab_select.setCurrentWidget(widget.mask_from_segmentation)
        assert widget.add_mask_btn.isEnabled() == (node_type is NodeType.segment)
        widget.mask_tab_select.setCurrentWidget(widget.mask_operation)
        assert widget.add_mask_btn.isEnabled() == (node_type is NodeType.root)

    def test_add_mask_segmentation(self, qtbot, part_settings):
        def check_mask(mask: prepare_plan_widget.MaskCreate):
            assert isinstance(mask, prepare_plan_widget.MaskCreate)
            assert mask.name == "mask_name"
            return True

        widget = prepare_plan_widget.SelectMaskOp(part_settings)
        qtbot.addWidget(widget)
        widget.mask_tab_select.setCurrentWidget(widget.mask_from_segmentation)
        widget.mask_name.setText("mask_name")
        with qtbot.waitSignal(widget.mask_step_add, check_params_cb=check_mask):
            widget._add_mask()

    def test_add_mask_file(self, qtbot, part_settings, tmp_path):
        file_path = str(tmp_path / "mask.txt")
        with open(file_path, "w") as f:
            f.write("mask")

        def check_mask(mask: prepare_plan_widget.MaskFile):
            assert isinstance(mask, prepare_plan_widget.MaskFile)
            assert mask.path_to_file == file_path
            return True

        widget = prepare_plan_widget.SelectMaskOp(part_settings)
        qtbot.addWidget(widget)
        widget.mask_tab_select.setCurrentWidget(widget.file_mask)
        with qtbot.waitSignal(widget.file_mask.select_type.currentEnumChanged):
            widget.file_mask.select_type.setCurrentEnum(prepare_plan_widget.FileMaskType.Mapping_file)
        widget.file_mask.first_text.setText(file_path)

        with qtbot.waitSignal(widget.mask_step_add, check_params_cb=check_mask):
            widget._add_mask()

    @pytest.mark.parametrize(
        ("enum", "klass"),
        [
            (prepare_plan_widget.MaskOperation.mask_intersection, prepare_plan_widget.MaskIntersection),
            (prepare_plan_widget.MaskOperation.mask_sum, prepare_plan_widget.MaskSum),
        ],
    )
    def test_add_mask_operation(self, qtbot, part_settings, enum, klass, monkeypatch):
        def check_mask(mask: klass):
            assert isinstance(mask, klass)
            assert mask.mask1 == "mask1"
            assert mask.mask2 == "mask2"
            return True

        widget = prepare_plan_widget.SelectMaskOp(part_settings)
        qtbot.addWidget(widget)
        widget.mask_tab_select.setCurrentWidget(widget.mask_operation)
        widget.mask_operation.setCurrentEnum(enum)
        monkeypatch.setattr(prepare_plan_widget.TwoMaskDialog, "exec_", lambda self: True)
        monkeypatch.setattr(prepare_plan_widget.TwoMaskDialog, "get_result", lambda self: ("mask1", "mask2"))

        with qtbot.waitSignal(widget.mask_step_add, check_params_cb=check_mask):
            widget._add_mask()


class TestCreatePlan:
    def test_create(self, qtbot, part_settings):
        widget = prepare_plan_widget.CreatePlan(part_settings)
        qtbot.addWidget(widget)

    def test_change_root_type(self, qtbot, part_settings, monkeypatch):
        monkeypatch.setattr(prepare_plan_widget.QInputDialog, "getText", lambda *args: ("root_type", True))

        widget = prepare_plan_widget.CreatePlan(part_settings)
        qtbot.addWidget(widget)
        widget.change_root_type(prepare_plan_widget.RootType.Project)
        widget.add_calculation_plan()
        assert "root_type" in part_settings.batch_plans
        assert part_settings.batch_plans["root_type"].execution_tree.operation == RootType.Project

    def test_add_roi_extraction(self, qtbot, part_settings, roi_extraction_profile):
        widget = prepare_plan_widget.CreatePlan(part_settings)
        qtbot.addWidget(widget)
        widget.add_roi_extraction(roi_extraction_profile)
        assert len(widget.calculation_plan.execution_tree.children) == 1
        assert len(widget.calculation_plan.execution_tree.children[0].children) == 0
        assert widget.calculation_plan.execution_tree.children[0].operation == roi_extraction_profile
        roi_extraction_profile2 = roi_extraction_profile.copy(update={"name": "roi_extraction_profile2"})
        widget.update_element_chk.setChecked(True)
        widget.add_roi_extraction(roi_extraction_profile2)
        assert len(widget.calculation_plan.execution_tree.children) == 1
        assert len(widget.calculation_plan.execution_tree.children[0].children) == 0
        assert widget.calculation_plan.execution_tree.children[0].operation == roi_extraction_profile2
        widget.clean_plan()
        assert len(widget.calculation_plan.execution_tree.children) == 0

    def test_add_save_operation(self, qtbot, part_settings):
        save_step_profile = Save(suffix="_save_step", directory="", algorithm="tiff", short_name="tiff", values={})
        widget = prepare_plan_widget.CreatePlan(part_settings)
        qtbot.addWidget(widget)
        widget.add_save_operation(save_step_profile)
        assert len(widget.calculation_plan.execution_tree.children) == 1
        assert len(widget.calculation_plan.execution_tree.children[0].children) == 0
        assert widget.calculation_plan.execution_tree.children[0].operation == save_step_profile
        save_step_profile2 = save_step_profile.copy(update={"suffix": "_save"})
        widget.update_element_chk.setChecked(True)
        widget.add_save_operation(save_step_profile2)
        assert len(widget.calculation_plan.execution_tree.children) == 1
        assert len(widget.calculation_plan.execution_tree.children[0].children) == 0
        assert widget.calculation_plan.execution_tree.children[0].operation == save_step_profile2
        widget.remove_element()
        assert len(widget.calculation_plan.execution_tree.children) == 0

    def test_add_set_of_measurement(self, qtbot, part_settings, measurement_profiles):
        widget = prepare_plan_widget.CreatePlan(part_settings)
        qtbot.addWidget(widget)
        measurement_calculate1 = MeasurementCalculate(
            channel=0, units=Units.nm, measurement_profile=measurement_profiles[0], name_prefix=""
        )
        measurement_calculate2 = MeasurementCalculate(
            channel=0, units=Units.nm, measurement_profile=measurement_profiles[1], name_prefix=""
        )
        widget.add_set_of_measurement(measurement_calculate1)
        assert len(widget.calculation_plan.execution_tree.children) == 1
        assert len(widget.calculation_plan.execution_tree.children[0].children) == 0
        assert widget.calculation_plan.execution_tree.children[0].operation == measurement_calculate1
        widget.update_element_chk.setChecked(True)
        widget.add_set_of_measurement(measurement_calculate2)
        assert len(widget.calculation_plan.execution_tree.children) == 1
        assert len(widget.calculation_plan.execution_tree.children[0].children) == 0
        assert widget.calculation_plan.execution_tree.children[0].operation == measurement_calculate2
        widget.clean_plan()
        assert len(widget.calculation_plan.execution_tree.children) == 0

    @patch("PartSeg._roi_analysis.prepare_plan_widget.show_warning")
    def test_add_roi_extraction_pipeline(
        self, show_warning_patch, qtbot, part_settings, roi_extraction_profile, mask_property
    ):
        segmentation_pipeline = SegmentationPipeline(
            name="test",
            segmentation=roi_extraction_profile,
            mask_history=[
                SegmentationPipelineElement(segmentation=roi_extraction_profile, mask_property=mask_property)
            ],
        )
        widget = prepare_plan_widget.CreatePlan(part_settings)
        qtbot.addWidget(widget)
        widget.add_roi_extraction_pipeline(segmentation_pipeline)
        assert len(widget.calculation_plan.execution_tree.children) == 1
        assert widget.calculation_plan.execution_tree.children[0].operation == roi_extraction_profile
        assert len(widget.calculation_plan.execution_tree.children[0].children) == 1
        assert widget.calculation_plan.execution_tree.children[0].children[0].operation.mask_property == mask_property
        assert len(widget.calculation_plan.execution_tree.children[0].children[0].children) == 1
        assert (
            widget.calculation_plan.execution_tree.children[0].children[0].children[0].operation
            == roi_extraction_profile
        )

        # test if update pipeline fail
        widget.update_element_chk.setChecked(True)
        show_warning_patch.assert_not_called()
        widget.add_roi_extraction_pipeline(segmentation_pipeline)
        show_warning_patch.assert_called_once_with("Cannot update pipeline", "Cannot update pipeline")

    @patch("PartSeg._roi_analysis.prepare_plan_widget.show_warning")
    def test_create_mask(
        self, show_warning_patch, qtbot, part_settings, mask_property, roi_extraction_profile, mask_property_non_default
    ):
        widget = prepare_plan_widget.CreatePlan(part_settings)
        qtbot.addWidget(widget)
        widget.add_roi_extraction(roi_extraction_profile)
        widget.create_mask(prepare_plan_widget.MaskCreate(name="test", mask_property=mask_property))
        assert len(widget.calculation_plan.execution_tree.children) == 1
        assert len(widget.calculation_plan.execution_tree.children[0].children) == 1
        assert widget.calculation_plan.execution_tree.children[0].children[0].operation.mask_property == mask_property
        assert widget.mask_set == {"test"}
        widget.update_element_chk.setChecked(True)
        widget.create_mask(prepare_plan_widget.MaskCreate(name="test2", mask_property=mask_property_non_default))
        assert (
            widget.calculation_plan.execution_tree.children[0].children[0].operation.mask_property
            == mask_property_non_default
        )
        assert widget.mask_set == {"test2"}

        # test adding mask with existing name
        show_warning_patch.assert_not_called()
        widget.create_mask(prepare_plan_widget.MaskCreate(name="test2", mask_property=mask_property_non_default))
        show_warning_patch.assert_called_once_with("Already exists", "Mask with this name already exists")

    def test_edit_plan(self, qtbot, part_settings, calculation_plan):
        widget = prepare_plan_widget.CreatePlan(part_settings)
        qtbot.addWidget(widget)
        assert count_tree_widget_items(widget.plan.topLevelItem(0)) == 1
        widget.edit_plan(calculation_plan)
        assert count_tree_widget_items(widget.plan.topLevelItem(0)) == 37

    @patch("PartSeg._roi_analysis.prepare_plan_widget.QMessageBox.warning")
    def test_edit_plan_bad_plan(self, message_patch, qtbot, part_settings, calculation_plan):
        widget = prepare_plan_widget.CreatePlan(part_settings)
        qtbot.addWidget(widget)
        calculation_plan.execution_tree.children[0].operation = {"__error__": NOT_FOUND_STR}
        widget.edit_plan(calculation_plan)
        message_patch.assert_called_once()


class TestCalculatePlaner:
    def test_create(self, qtbot, part_settings):
        widget = prepare_plan_widget.CalculatePlaner(part_settings)
        qtbot.addWidget(widget)


class TestPlanPreview:
    def test_create(self, qtbot, calculation_plan):
        widget = prepare_plan_widget.PlanPreview(calculation_plan=calculation_plan)
        qtbot.addWidget(widget)

    @patch("PartSeg._roi_analysis.prepare_plan_widget.QMessageBox.warning")
    def test_create_bad_plan(self, message_patch, qtbot, calculation_plan):
        widget = prepare_plan_widget.PlanPreview()
        qtbot.addWidget(widget)
        calculation_plan.execution_tree.children[0].operation = {"__error__": NOT_FOUND_STR}
        widget.set_plan(calculation_plan)
        message_patch.assert_called_once()
        assert widget.calculation_plan is None


def test_calculation_plan_repr(calculation_plan):
    assert "name='test'" in repr(calculation_plan)
    assert "operation=<RootType.Image: 0>" in repr(calculation_plan)


def count_tree_widget_items(tree_widget: QTreeWidgetItem):
    count = 1
    for i in range(tree_widget.childCount()):
        count += count_tree_widget_items(tree_widget.child(i))
        count += 1

    return count



import warnings
from abc import ABC

import numpy as np
import SimpleITK as sitk
from local_migrator import update_argument
from pydantic import Field

from PartSegCore.algorithm_describe_base import AlgorithmDescribeBase, AlgorithmSelection
from PartSegCore.segmentation.watershed import NeighType, get_neighbourhood
from PartSegCore.utils import BaseModel


class BaseSmoothing(AlgorithmDescribeBase, ABC):
    __argument_class__ = BaseModel

    @classmethod
    def smooth(cls, segmentation: np.ndarray, arguments: dict) -> np.ndarray:
        raise NotImplementedError


class NoneSmoothing(BaseSmoothing):
    @classmethod
    def get_name(cls) -> str:
        return "None"

    @classmethod
    def smooth(cls, segmentation: np.ndarray, arguments: dict) -> np.ndarray:
        return segmentation


class OpeningSmoothingParams(BaseModel):
    smooth_border_radius: int = Field(2, title="Smooth borders radius", ge=1, le=20)


class OpeningSmoothing(BaseSmoothing):
    __argument_class__ = OpeningSmoothingParams

    @classmethod
    def get_name(cls) -> str:
        return "Opening"

    @classmethod
    @update_argument("arguments")
    def smooth(cls, segmentation: np.ndarray, arguments: OpeningSmoothingParams) -> np.ndarray:
        radius = arguments.smooth_border_radius
        if isinstance(radius, (int, float)):
            radius = [radius] * segmentation.ndim
        return sitk.GetArrayFromImage(sitk.BinaryMorphologicalOpening(sitk.GetImageFromArray(segmentation), radius))


class VoteSmoothingParams(BaseModel):
    neighbourhood_type: NeighType = Field(
        NeighType.edges, title="Side Neighbourhood", description="use 6, 18 or 26 neighbourhood (5, 8, 8 for 2d data)"
    )
    support_level: int = Field(
        1,
        title="Support level",
        ge=1,
        le=27,
        description="How many voxels in neighbourhood need to be labeled to preserve pixel",
    )


class VoteSmoothing(BaseSmoothing):
    __argument_class__ = VoteSmoothingParams

    @classmethod
    def get_name(cls) -> str:
        return "Vote"

    @classmethod
    @update_argument("arguments")
    def smooth(cls, segmentation: np.ndarray, arguments: VoteSmoothingParams) -> np.ndarray:
        segmentation_bin = (segmentation > 0).astype(np.uint8)
        count_array = np.zeros(segmentation_bin.shape, dtype=segmentation.dtype)
        neighbourhood = get_neighbourhood(segmentation_bin.squeeze().shape, arguments.neighbourhood_type)
        axis = tuple(range(len(segmentation_bin.shape)))
        for shift in neighbourhood:
            count_array += np.roll(segmentation_bin, shift[: segmentation_bin.ndim], axis)
        segmentation = segmentation.copy()
        count_array = count_array.reshape(segmentation.shape)
        segmentation[count_array < arguments.support_level] = 0
        return segmentation


class IterativeSmoothingParams(VoteSmoothingParams):
    max_steps: int = Field(
        1, title="Max steps", description="How many voxels in neighbourhood need to be labeled to preserve pixel"
    )


class IterativeVoteSmoothing(BaseSmoothing):
    __argument_class__ = IterativeSmoothingParams

    @classmethod
    def get_name(cls) -> str:
        return "Iterative Vote"

    @classmethod
    @update_argument("arguments")
    def smooth(cls, segmentation: np.ndarray, arguments: IterativeSmoothingParams) -> np.ndarray:
        segmentation_bin = (segmentation > 0).astype(np.uint8)
        count_array = np.zeros(segmentation_bin.shape, dtype=segmentation.dtype)
        neighbourhood = get_neighbourhood(segmentation_bin.squeeze().shape, arguments.neighbourhood_type)
        segmentation = segmentation.copy()
        count_point = np.count_nonzero(segmentation)
        axis = tuple(range(len(segmentation_bin.shape)))
        for _ in range(arguments.max_steps):
            for shift in neighbourhood:
                count_array += np.roll(segmentation_bin, shift[: segmentation_bin.ndim], axis)
            segmentation_bin[count_array < arguments.support_level] = 0
            count_point2 = np.count_nonzero(segmentation_bin)
            if count_point2 == count_point:
                break
            count_point = count_point2
            count_array[:] = 0
        segmentation[segmentation_bin.reshape(segmentation.shape) == 0] = 0
        return segmentation


class SmoothAlgorithmSelection(AlgorithmSelection, class_methods=["smooth"], suggested_base_class=BaseSmoothing):
    pass


SmoothAlgorithmSelection.register(NoneSmoothing)
SmoothAlgorithmSelection.register(OpeningSmoothing)
SmoothAlgorithmSelection.register(VoteSmoothing)
SmoothAlgorithmSelection.register(IterativeVoteSmoothing)


def __getattr__(name):  # pragma: no cover
    if name == "smooth_dict":
        warnings.warn(
            "threshold_dict is deprecated. Please use SmoothAlgorithmSelection instead",
            category=FutureWarning,
            stacklevel=2,
        )
        return SmoothAlgorithmSelection.__register__

    raise AttributeError(f"module {__name__} has no attribute {name}")



import logging
import sys

import sentry_sdk
from qtpy.QtCore import QCoreApplication, QThread
from qtpy.QtWidgets import QMessageBox
from superqt import ensure_main_thread

from PartSeg import state_store
from PartSegCore.segmentation.algorithm_base import SegmentationLimitException
from PartSegImage import TiffFileException


def my_excepthook(type_, value, trace_back):
    """
    Custom excepthook.
    Close application on :py:class:`KeyboardInterrupt`.

    If :py:data:`PartSeg.state_store.always_report` is set then just sent report using sentry.
    otherwise show dialog with information about error and ask user
    if he wants to send report using :py:func:`show_error`.
    """

    # log the exception here
    if state_store.show_error_dialog and not isinstance(value, KeyboardInterrupt):
        if state_store.auto_report or state_store.always_report:
            with sentry_sdk.push_scope() as scope:
                scope.set_tag("auto_report", "true")
                scope.set_tag("main_thread", QCoreApplication.instance().thread() == QThread.currentThread())
                sentry_sdk.capture_exception(value)
        if state_store.always_report:
            return
        try:
            show_error(value)
        except ImportError:
            sys.__excepthook__(type_, value, trace_back)
    elif isinstance(value, KeyboardInterrupt):
        logging.warning("KeyboardInterrupt close")
        sys.exit(1)
    else:
        # then call the default handler
        sys.__excepthook__(type_, value, trace_back)


@ensure_main_thread
def show_error(error=None):
    """
    For :py:class:`SegmentationLimitException` and :py:class:`TiffFileException`
    show dialog with information about problem.

    For other exceptions show :py:class:`ErrorDialog` dialog
    with information about error that allow to report it.

    :param error: exception to show
    """
    if error is None:
        return

    from PartSeg.common_gui.error_report import ErrorDialog, QMessageFromException

    if isinstance(error, TiffFileException):
        QMessageFromException.critical(
            None,
            "Tiff file error",
            f"During read file there is an error: {', '.join(str(x) for x in error.args)}",
            exception=error,
        )
        return
    if isinstance(error, SegmentationLimitException):
        QMessageFromException.critical(
            None,
            "Segmentation limitations",
            f"During segmentation process algorithm meet limitations: {', '.join(str(x) for x in error.args)}",
            exception=error,
        )
        return

    dial = ErrorDialog(error, "Exception during program run")
    dial.exec_()


@ensure_main_thread
def show_warning(header=None, text=None, exception=None):
    """
    Show warning :py:class:`PyQt5.QtWidgets.QMessageBox`

    This function is to ensure creation warning dialog in main thread.
    """
    if exception is not None:
        from PartSeg.common_gui.error_report import QMessageFromException

        message = QMessageFromException(QMessageBox.Icon.Warning, header, text, exception=exception)
    else:
        message = QMessageBox(QMessageBox.Icon.Warning, header, text, QMessageBox.StandardButton.Ok)
    message.exec_()



import inspect
import itertools
import typing

from magicgui.widgets import Widget, create_widget
from napari import Viewer
from napari.layers import Image as NapariImage
from napari.layers import Labels, Layer
from qtpy.QtWidgets import QWidget

from PartSeg.common_gui.algorithms_description import FormWidget, QtAlgorithmProperty
from PartSeg.common_gui.custom_save_dialog import FormDialog
from PartSegCore import UNIT_SCALE, Units
from PartSegCore.algorithm_describe_base import AlgorithmProperty
from PartSegImage import Channel, Image


class QtNapariAlgorithmProperty(QtAlgorithmProperty):
    @classmethod
    def _get_field_from_value_type(cls, ap: AlgorithmProperty) -> typing.Union[QWidget, Widget]:
        if inspect.isclass(ap.value_type) and issubclass(ap.value_type, Channel):
            return create_widget(annotation=NapariImage, label="Image", options={})
        return super()._get_field_from_value_type(ap)


class NapariFormWidget(FormWidget):
    @staticmethod
    def _element_list_map(fields) -> typing.Iterable[QtAlgorithmProperty]:
        return map(QtNapariAlgorithmProperty.from_algorithm_property, fields)

    def reset_choices(self, event=None):
        for widget in self.widgets_dict.values():
            if hasattr(widget.get_field(), "reset_choices"):
                widget.get_field().reset_choices(event)


class NapariFormWidgetWithMask(NapariFormWidget):
    @classmethod
    def _element_list_map(cls, fields) -> typing.Iterable[QtAlgorithmProperty]:
        mask = AlgorithmProperty("mask", "Mask", None, value_type=typing.Optional[Labels])
        return super()._element_list_map(itertools.chain([mask], fields))

    def get_layers(self):
        return {name: el.get_value() for name, el in self.widgets_dict.items() if isinstance(el.get_value(), Layer)}

    def get_values(self):
        res = {name: el.get_value() for name, el in self.widgets_dict.items() if name != "mask"}
        return self._model_class(**res) if self._model_class is not None else res


class NapariFormDialog(FormDialog):
    @staticmethod
    def widget_class() -> typing.Type[FormWidget]:
        return NapariFormWidget

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.widget.reset_choices()


def generate_image(viewer: Viewer, *layer_names):
    axis_order = Image.axis_order.replace("C", "")
    image_list = []
    for name in dict.fromkeys(layer_names):
        image_layer = viewer.layers[name]
        data_scale = image_layer.scale[-3:] / UNIT_SCALE[Units.nm.value]
        image_list.append(
            Image(
                image_layer.data,
                data_scale,
                axes_order=axis_order[-image_layer.data.ndim :],
                channel_names=[image_layer.name],
            )
        )
    res_image = image_list[0]
    for image in image_list[1:]:
        res_image = res_image.merge(image, "C")

    return res_image



import qtawesome as qta
from qtpy.QtGui import QPainter, QPaintEvent
from qtpy.QtWidgets import QCheckBox


class LockCheckBox(QCheckBox):
    def paintEvent(self, event: QPaintEvent):
        rect = event.rect()
        painter = QPainter(self)
        color = painter.pen().color()
        lock = qta.icon("fa5s.lock", color=color) if self.isChecked() else qta.icon("fa5s.lock-open", color=color)
        lock.paint(painter, rect)



import itertools
import logging
from contextlib import suppress
from dataclasses import dataclass, field
from enum import Enum
from functools import partial
from typing import TYPE_CHECKING, Dict, List, MutableMapping, Optional, Tuple, Union

import napari
import numpy as np
from local_migrator import register_class
from napari.components import ViewerModel as Viewer
from napari.layers import Layer, Points
from napari.layers.image import Image as NapariImage
from napari.layers.labels import Labels
from napari.qt import QtViewer
from napari.qt.threading import thread_worker
from napari.utils.colormaps.colormap import ColormapInterpolationMode
from packaging.version import parse as parse_version
from qtpy.QtCore import QEvent, QPoint, Qt, QTimer, Signal, Slot
from qtpy.QtWidgets import QApplication, QCheckBox, QHBoxLayout, QLabel, QMenu, QSpinBox, QToolTip, QVBoxLayout, QWidget
from scipy.ndimage import binary_dilation
from superqt import QEnumComboBox, ensure_main_thread
from vispy.color import Color, Colormap
from vispy.geometry.rect import Rect

from PartSeg.common_backend.base_settings import BaseSettings
from PartSeg.common_gui.advanced_tabs import RENDERING_LIST, RENDERING_MODE_NAME_STR, SEARCH_ZOOM_FACTOR_STR
from PartSeg.common_gui.channel_control import ChannelProperty, ColorComboBoxGroup
from PartSeg.common_gui.custom_buttons import SearchROIButton
from PartSeg.common_gui.qt_modal import QtPopup
from PartSegCore.image_operations import NoiseFilterType, bilateral, gaussian, median
from PartSegCore.roi_info import ROIInfo
from PartSegImage import Image

if TYPE_CHECKING:
    from vispy.scene import BaseCamera

try:
    from napari._qt.qt_viewer_buttons import QtViewerPushButton as QtViewerPushButton_
except ImportError:
    from napari._qt.widgets.qt_viewer_buttons import QtViewerPushButton as QtViewerPushButton_
_napari_ge_4_13 = parse_version(napari.__version__) >= parse_version("0.4.13a1")
_napari_ge_4_17 = parse_version(napari.__version__) >= parse_version("0.4.17a1")
_napari_ge_5 = parse_version(napari.__version__) >= parse_version("0.5.0a1")


def get_highlight_colormap():
    cmap_dict = {0: (0, 0, 0, 0), 1: "white", None: (0, 0, 0, 0)}
    if _napari_ge_5:
        from napari.utils.colormaps import DirectLabelColormap

        return {"colormap": DirectLabelColormap(color_dict=cmap_dict)}

    return {"color": cmap_dict}


class QtViewerPushButton(QtViewerPushButton_):
    def __init__(self, viewer, *args, **kwargs):
        if _napari_ge_4_13:
            super().__init__(*args, **kwargs)
        else:
            super().__init__(viewer, *args, **kwargs)


if _napari_ge_4_13:

    class QtNDisplayButton(QtViewerPushButton_):
        def __init__(self, viewer):
            super().__init__(button_name="ndisplay_button", tooltip="Toggle dimensions", slot=self.toggle_ndisplay)
            self.viewer = viewer
            self.setCheckable(True)

        def toggle_ndisplay(self):
            self.viewer.dims.ndisplay = 2 + (self.viewer.dims.ndisplay == 2)

else:
    from napari.qt import QtStateButton

    class QtNDisplayButton(QtStateButton):
        def __init__(self, viewer):
            super().__init__(
                "ndisplay_button",
                viewer.dims,
                "ndisplay",
                viewer.dims.events.ndisplay,
                2,
                3,
            )


ORDER_DICT = {"xy": [0, 1, 2, 3], "zy": [0, 2, 1, 3], "zx": [0, 3, 1, 2]}
NEXT_ORDER = {"xy": "zy", "zy": "zx", "zx": "xy"}

ColorInfo = Dict[int, Union[str, List[float]]]


@dataclass
class ImageInfo:
    image: Image
    layers: List[NapariImage]
    filter_info: List[Tuple[NoiseFilterType, float]] = field(default_factory=list)
    mask: Optional[Labels] = None
    mask_array: Optional[np.ndarray] = None
    roi: Optional[Labels] = None
    roi_info: ROIInfo = field(default_factory=lambda: ROIInfo(None))
    roi_count: int = 0
    highlight: Optional[Labels] = None

    def coords_in(self, coords: Union[List[int], np.ndarray]) -> bool:
        if not self.layers:
            return False
        fst_layer = self.layers[0]
        moved_coords = self.translated_coords(coords)
        return np.all(moved_coords >= 0) and np.all(moved_coords < fst_layer.data.shape)

    def translated_coords(self, coords: Union[List[int], np.ndarray]) -> np.ndarray:
        if not self.layers:
            return np.array(coords)
        fst_layer = self.layers[0]
        return np.subtract(coords, fst_layer.translate).astype(int)


@register_class(old_paths=["PartSeg.common_gui.stack_image_view.LabelEnum"])
class LabelEnum(Enum):
    Not_show = 0
    Show_results = 1
    Show_selected = 2

    def __str__(self):
        return "Don't show" if self.value == 0 else self.name.replace("_", " ")


class SearchType(Enum):
    Highlight = 0
    Zoom_in = 1


class ImageView(QWidget):
    position_changed = Signal([int, int, int], [int, int])
    component_clicked = Signal(int)
    text_info_change = Signal(str)
    hide_signal = Signal(bool)
    view_changed = Signal()
    image_added = Signal()

    def __init__(
        self,
        settings: BaseSettings,
        channel_property: ChannelProperty,
        name: str,
        parent: Optional[QWidget] = None,
        ndisplay=2,
    ):
        super().__init__(parent=parent)

        self.settings = settings
        self.channel_property = channel_property
        self.name = name
        self.image_info: Dict[str, ImageInfo] = {}
        self.current_image = ""
        self._current_order = "xy"
        self.components = None
        self.worker_list = []
        self.points_layer = None
        self.roi_alternative_selection = "ROI"
        self._search_type = SearchType.Highlight
        self._last_component = 1

        self.viewer = Viewer(ndisplay=ndisplay)
        self.viewer.theme = self.settings.theme_name
        self.viewer_widget = NapariQtViewer(self.viewer)

        self._update_scale_bar_ticks()

        self.channel_control = ColorComboBoxGroup(settings, name, channel_property, height=30)
        self.ndim_btn = QtNDisplayButton(self.viewer)
        self.reset_view_button = QtViewerPushButton(self.viewer, "home", "Reset view", self._reset_view)
        self.points_view_button = QtViewerPushButton(
            self.viewer, "new_points", "Show points", self.toggle_points_visibility
        )
        self.points_view_button.setVisible(False)
        self.search_roi_btn = SearchROIButton(self.settings)
        self.search_roi_btn.clicked.connect(self._search_component)
        self.search_roi_btn.setDisabled(True)
        self.roll_dim_button = QtViewerPushButton(self.viewer, "roll", "Roll dimension", self._rotate_dim)
        self.roll_dim_button.setContextMenuPolicy(Qt.ContextMenuPolicy.CustomContextMenu)
        self.roll_dim_button.customContextMenuRequested.connect(self._dim_order_menu)
        self.mask_chk = QCheckBox()
        self.mask_chk.setVisible(False)
        self.mask_label = QLabel("Mask:")
        self.mask_label.setVisible(False)

        self.btn_layout = QHBoxLayout()
        self.btn_layout2 = QHBoxLayout()
        self.setup_ui()

        self.channel_control.change_channel.connect(self.change_visibility)
        self.viewer.events.status.connect(self.print_info)

        self._connect_to_settings()

        self.old_scene: BaseCamera = self.viewer_widget.view.scene

        self.mask_chk.stateChanged.connect(self.change_mask_visibility)
        self.viewer_widget.view.scene.transform.changed.connect(self._view_changed, position="last")
        self.viewer.dims.events.current_step.connect(self._view_changed, position="last")
        self.viewer.dims.events.ndisplay.connect(self._view_changed, position="last")
        self.viewer.dims.events.ndisplay.connect(self._view_changed, position="last")
        self.viewer.dims.events.ndisplay.connect(self.camera_change, position="last")
        self.viewer.events.reset_view.connect(self._view_changed, position="last")

    def setup_ui(self):
        self.btn_layout.addWidget(self.reset_view_button)
        self.btn_layout.addWidget(self.ndim_btn)
        self.btn_layout.addWidget(self.roll_dim_button)
        self.btn_layout.addWidget(self.points_view_button)
        self.btn_layout.addWidget(self.search_roi_btn)
        self.btn_layout.addWidget(self.channel_control, 1)
        self.btn_layout.addWidget(self.mask_label)
        self.btn_layout.addWidget(self.mask_chk)
        layout = QVBoxLayout()
        layout.addLayout(self.btn_layout)
        layout.addLayout(self.btn_layout2)
        layout.addWidget(self.viewer_widget)

        self.setLayout(layout)

        if hasattr(self.viewer_widget.canvas, "background_color_override"):
            self.viewer_widget.canvas.background_color_override = "black"
            if _napari_ge_4_17:
                self.viewer.scale_bar.color = "white"
                self.viewer.scale_bar.colored = True

    def _connect_to_settings(self):
        self.settings.mask_changed.connect(self.set_mask)
        self.settings.roi_changed.connect(self.set_roi)
        self.settings.roi_clean.connect(self.set_roi)
        self.settings.image_changed.connect(self.set_image)
        self.settings.image_spacing_changed.connect(self.update_spacing_info)
        self.settings.points_changed.connect(self.update_points)
        self.settings.connect_to_profile(RENDERING_MODE_NAME_STR, self.update_rendering)
        self.settings.labels_changed.connect(self.update_roi_coloring)
        self.settings.connect_to_profile(f"{self.name}.image_state.opacity", self.update_roi_coloring)
        self.settings.connect_to_profile(f"{self.name}.image_state.only_border", self.update_roi_border)
        self.settings.connect_to_profile(f"{self.name}.image_state.border_thick", self.update_roi_border)
        self.settings.connect_to_profile(f"{self.name}.image_state.show_label", self.update_roi_labeling)
        self.settings.connect_to_profile("mask_presentation_opacity", self.update_mask_parameters)
        self.settings.connect_to_profile("mask_presentation_color", self.update_mask_parameters)
        self.settings.connect_to_profile("scale_bar_ticks", self._update_scale_bar_ticks)

    def _update_scale_bar_ticks(self):
        self.viewer.scale_bar.ticks = self.settings.get_from_profile("scale_bar_ticks", True)

    def toggle_points_visibility(self):
        if self.points_layer is not None:
            self.points_layer.visible = not self.points_layer.visible

    def toggle_scale_bar(self):
        self.viewer.scale_bar.unit = "nm"
        self.viewer.scale_bar.visible = not self.viewer.scale_bar.visible

    def _dim_order_menu(self, point: QPoint):
        menu = QMenu()
        for key in ORDER_DICT:
            action = menu.addAction(key)
            action.triggered.connect(partial(self._set_new_order, key))
            if key == self._current_order:
                font = action.font()
                font.setBold(True)
                action.setFont(font)

        menu.exec_(self.roll_dim_button.mapToGlobal(point))

    def _set_new_order(self, text: str):
        self._current_order = text
        self.viewer.dims.order = ORDER_DICT[text]

    def _reset_view(self):
        self._set_new_order("xy")
        self.viewer.dims.order = ORDER_DICT[self._current_order]
        self.viewer.reset_view()

    def _rotate_dim(self):
        self._set_new_order(NEXT_ORDER[self._current_order])

    def camera_change(self, _args):
        self.old_scene.transform.changed.disconnect(self._view_changed)
        self.old_scene: BaseCamera = self.viewer_widget.view.camera
        self.old_scene.transform.changed.connect(self._view_changed, position="last")

    def _view_changed(self, _args):
        self.view_changed.emit()

    def get_state(self):
        return {
            "ndisplay": self.viewer.dims.ndisplay,
            "point": self.viewer.dims.point,
            "camera": self.viewer_widget.view.camera.get_state(),
        }

    def set_state(self, dkt):
        if "ndisplay" in dkt and self.viewer.dims.ndisplay != dkt["ndisplay"]:
            self.viewer.dims.ndisplay = dkt["ndisplay"]
            return
        if "point" in dkt:
            for i, val in enumerate(dkt["point"]):
                self.viewer.dims.set_point(i, val)
        if "camera" in dkt:
            with suppress(KeyError):
                self.viewer_widget.view.camera.set_state(dkt["camera"])

    def change_mask_visibility(self):
        for image_info in self.image_info.values():
            if image_info.mask is not None:
                image_info.mask.visible = self.mask_chk.isChecked()

    def update_spacing_info(self, image: Optional[Image] = None) -> None:
        """
        Update spacing of image if not provide, then use image pointed by settings.

        :param Optional[Image] image: image which spacing should be updated.
        :return: None
        """
        if image is None:
            image = self.settings.image

        if image.file_path not in self.image_info:
            raise ValueError("Image not registered")

        image_info = self.image_info[image.file_path]

        for layer in image_info.layers:
            layer.scale = image.normalized_scaling()

        if image_info.roi is not None:
            image_info.roi.scale = image.normalized_scaling()

        if image_info.mask is not None:
            image_info.mask.scale = image.normalized_scaling()

    def _active_layer(self):
        if hasattr(self.viewer.layers, "selection"):
            return self.viewer.layers.selection.active
        return self.viewer.active_layer

    def _coordinates(self):
        active_layer = self._active_layer()
        if active_layer is None:
            return None
        if (
            hasattr(self.viewer, "cursor")
            and hasattr(self.viewer.cursor, "position")
            and hasattr(active_layer, "world_to_data")
        ):
            return [int(x) for x in active_layer.world_to_data(self.viewer.cursor.position)]
        return [int(x) for x in active_layer.coordinates]

    def print_info(self, event=None):
        cords = self._coordinates()
        if cords is None:
            return
        bright_array = []
        components = []
        alt_components = []
        for image_info in self.image_info.values():
            if not image_info.coords_in(cords):
                continue
            moved_coords = image_info.translated_coords(cords)
            bright_array.extend(layer.data[tuple(moved_coords)] for layer in image_info.layers if layer.visible)

            if (
                image_info.roi_info.roi is not None
                and image_info.roi is not None
                and (val := image_info.roi_info.roi[tuple(moved_coords)])
            ):
                components.append(val)
            if self.roi_alternative_selection in image_info.roi_info.alternative and (
                val := image_info.roi_info.alternative[self.roi_alternative_selection][tuple(moved_coords)]
            ):
                alt_components.append(val)

        if not bright_array and not components:
            self.text_info_change.emit("")
            return
        text = f"{cords}: "
        if bright_array:
            text += str(bright_array[0]) if len(bright_array) == 1 else str(bright_array)
        self.components = components
        text += _print_list(components, " component")
        text += _print_list(alt_components, " alt")
        self.text_info_change.emit(text)

    def mask_opacity(self) -> float:
        """Get mask opacity"""
        return self.settings.get_from_profile("mask_presentation_opacity", 1)

    def mask_color(self) -> ColorInfo:
        """Get mask marking color"""
        color = Color(np.divide(self.settings.get_from_profile("mask_presentation_color", [255, 255, 255]), 255))
        return {0: (0, 0, 0, 0), 1: color.rgba}

    def get_image(self, image: Optional[Image]) -> Image:
        if image is not None:
            return image
        if self.current_image not in self.image_info:
            return self.settings.image
        return self.image_info[self.current_image].image

    def update_points(self):
        if self.settings.points is not None:
            self.points_view_button.setVisible(True)
            if self.points_layer is None or self.points_layer not in self.viewer.layers:
                self.points_layer = Points(self.settings.points, scale=self.settings.image.normalized_scaling())
                self.viewer.add_layer(self.points_layer)
            else:
                self.points_layer.data = self.settings.points
                self.points_layer.scale = self.settings.image.normalized_scaling()
        elif self.points_layer is not None and self.points_layer in self.viewer.layers:
            self.points_view_button.setVisible(False)
            self.points_layer.data = np.empty((0, 4))

    def set_roi(self, roi_info: Optional[ROIInfo] = None, image: Optional[Image] = None) -> None:
        image = self.get_image(image)
        if roi_info is None:
            roi_info = self.settings.roi_info
        if image.file_path not in self.image_info:
            return
        image_info = self.image_info[image.file_path]
        if image_info.roi is None and roi_info.roi is not None:
            image_info.roi_info = roi_info
            self.add_roi_layer(image_info)
            self.search_roi_btn.setDisabled(False)
        elif image_info.roi is None:
            return
        elif roi_info.roi is None:
            image_info.roi.visible = False
            self.search_roi_btn.setDisabled(True)
            return
        else:
            image_info.roi_info = roi_info
            image_info.roi.data = roi_info.alternative.get(self.roi_alternative_selection, roi_info.roi)
            image_info.roi.visible = True
            self.search_roi_btn.setDisabled(False)

        image_info.roi_count = max(roi_info.bound_info) if roi_info.bound_info else 0

        if image_info.roi not in self.viewer.layers:
            self.viewer.add_layer(image_info.roi)

        self.set_roi_colormap(image_info)
        image_info.roi.opacity = self.settings.get_from_profile(f"{self.name}.image_state.opacity", 1.0)
        image_info.roi.refresh()

    def get_roi_view_parameters(self, image_info: ImageInfo) -> ColorInfo:
        colors = self.settings.label_colors / 255
        if (
            self.settings.get_from_profile(f"{self.name}.image_state.show_label", LabelEnum.Show_results)
            == LabelEnum.Not_show
            or image_info.roi_count == 0
            or colors.size == 0
        ):
            return {x: [0, 0, 0, 0] for x in range(image_info.roi_count + 1)}

        res = {x: colors[(x - 1) % colors.shape[0]] for x in range(1, image_info.roi_count + 1)}
        res[0] = [0, 0, 0, 0]
        return res

    def set_roi_colormap(self, image_info) -> None:
        if _napari_ge_4_13:
            image_info.roi.color = self.get_roi_view_parameters(image_info)
            return
        colors = self.settings.label_colors / 255
        if (
            self.settings.get_from_profile(f"{self.name}.image_state.show_label", LabelEnum.Show_results)
            == LabelEnum.Not_show
            or image_info.roi_count == 0
            or colors.size == 0
        ):
            image_info.roi.colormap = Colormap([[0, 0, 0, 0], [0, 0, 0, 0]])

        res = [[*list(colors[(x - 1) % colors.shape[0]]), 1] for x in range(image_info.roi_count + 1)]
        res[0] = [0, 0, 0, 0]
        if len(res) < 2:
            res += [[0, 0, 0, 0] for _ in range(2 - len(res))]

        image_info.roi.colormap = Colormap(colors=res, interpolation=ColormapInterpolationMode.ZERO)
        max_val = image_info.roi_count + 1
        image_info.roi._all_vals = np.array(  # pylint: disable=protected-access
            [0] + [(x + 1) / (max_val + 1) for x in range(1, max_val)]
        )

    def update_roi_coloring(self):
        for image_info in self.image_info.values():
            if image_info.roi is None:
                continue
            self.set_roi_colormap(image_info)
            image_info.roi.opacity = self.settings.get_from_profile(f"{self.name}.image_state.opacity", 1.0)

    def remove_all_roi(self):
        for image_info in self.image_info.values():
            if image_info.roi is None:
                continue
            image_info.roi.visible = False

    def update_roi_border(self) -> None:
        for image_info in self.image_info.values():
            if image_info.roi is None:
                continue
            roi = image_info.roi_info.alternative.get(self.roi_alternative_selection, image_info.roi_info.roi)
            border_thick = self.settings.get_from_profile(f"{self.name}.image_state.border_thick", 1)
            only_border = self.settings.get_from_profile(f"{self.name}.image_state.only_border", True)
            alternative = image_info.roi.metadata.get("alternative", self.roi_alternative_selection)
            if alternative != self.roi_alternative_selection:
                image_info.roi.data = roi
            image_info.roi.contour = border_thick if only_border else 0
            image_info.roi.metadata["alternative"] = self.roi_alternative_selection

    @ensure_main_thread
    def update_rendering(self):
        rendering = self.settings.get_from_profile(RENDERING_MODE_NAME_STR, RENDERING_LIST[0])
        for image_info in self.image_info.values():
            if image_info.roi is not None and hasattr(image_info.roi, "rendering"):
                image_info.roi.rendering = rendering

    @ensure_main_thread
    def update_roi_labeling(self):
        for image_info in self.image_info.values():
            if image_info.roi is not None:
                self.set_roi_colormap(image_info)

    def add_roi_layer(self, image_info: ImageInfo):
        if image_info.roi_info.roi is None:
            return
        roi = image_info.roi_info.alternative.get(self.roi_alternative_selection, image_info.roi_info.roi)
        border_thick = self.settings.get_from_profile(f"{self.name}.image_state.border_thick", 1)
        kwargs = {
            "scale": image_info.image.normalized_scaling(),
            "name": "ROI",
            "blending": "translucent",
            "metadata": {"alternative": self.roi_alternative_selection},
            "rendering": self.settings.get_from_profile(RENDERING_MODE_NAME_STR, RENDERING_LIST[0]),
        }

        only_border = self.settings.get_from_profile(f"{self.name}.image_state.only_border", True)
        image_info.roi = self.viewer.add_labels(roi, **kwargs)
        image_info.roi.contour = border_thick if only_border else 0

    def set_mask(self, mask: Optional[np.ndarray] = None, image: Optional[Image] = None) -> None:
        image = self.get_image(image)
        if image.file_path not in self.image_info:
            raise ValueError("Image not added to viewer")
        if mask is None:
            mask = image.mask

        image_info = self.image_info[image.file_path]
        if mask is None:
            if image_info.mask is not None:
                image_info.mask.visible = False
                image_info.mask.metadata["valid"] = False
                self._toggle_mask_chk_visibility()
            return
        mask_marker = mask == 0
        if image_info.mask is None:
            image_info.mask = self.viewer.add_labels(
                mask_marker, scale=image.normalized_scaling(), blending="translucent", name="Mask"
            )
        else:
            image_info.mask.data = mask_marker
        image_info.mask.metadata["valid"] = True
        image_info.mask.color = self.mask_color()
        image_info.mask.opacity = self.mask_opacity()
        image_info.mask.visible = self.mask_chk.isChecked()
        self._toggle_mask_chk_visibility()

    def _toggle_mask_chk_visibility(self):
        visibility = any(
            image_info.mask is not None and image_info.mask.metadata.get("valid", False)
            for image_info in self.image_info.values()
        )
        self.mask_chk.setVisible(visibility)
        self.mask_label.setVisible(visibility)

    def update_mask_parameters(self):
        opacity = self.mask_opacity()
        colormap = self.mask_color()
        for image_info in self.image_info.values():
            if image_info.mask is not None:
                image_info.mask.opacity = opacity
                image_info.mask.color = colormap

    def set_image(self, image: Optional[Image] = None):
        self.image_info = {}
        self.add_image(image, True)

    def has_image(self, image: Image):
        return image.file_path in self.image_info

    @staticmethod
    def calculate_filter(array: np.ndarray, parameters: Tuple[NoiseFilterType, float]) -> Optional[np.ndarray]:
        if parameters[0] == NoiseFilterType.No or parameters[1] == 0:
            return array
        if parameters[0] == NoiseFilterType.Gauss:
            return gaussian(array, parameters[1])
        if parameters[0] == NoiseFilterType.Bilateral:
            return bilateral(array, parameters[1])
        return median(array, int(parameters[1]))

    def _remove_worker(self, sender=None):
        if sender is None:
            sender = self.sender()
        for worker in self.worker_list:
            if hasattr(worker, "signals") and sender is worker.signals:
                self.worker_list.remove(worker)
                break
        else:
            logging.debug("[_remove_worker] %s", sender)

    def _add_layer_util(self, index, layer, filters):
        if layer not in self.viewer.layers:
            self.viewer.add_layer(layer)

        if filters[index][0] == NoiseFilterType.No or filters[index][1] == 0:
            return

        worker = calc_layer_filter(layer, filters[index][0], filters[index][1])
        worker.returned.connect(self._add_layer_util_end)
        worker.finished.connect(self._remove_worker)
        self.worker_list.append(worker)
        worker.start()

    @Slot(object)
    def _add_layer_util_end(self, val):
        data_, layer_ = val
        if data_ is not None:
            layer_.data = data_

    def _add_or_move_layer(self, layer: Optional[Layer], index):
        if layer is None:
            return
        if layer not in self.viewer.layers:
            self.viewer.add_layer(layer)
        else:
            self.viewer.layers.move(self.viewer.layers.index(layer), index)

    def _add_image(self, image_data: Tuple[ImageInfo, bool]):
        image_info, replace = image_data
        image = image_info.image

        if replace:
            for layer in list(reversed(self.viewer.layers)):
                self.viewer.layers.remove(layer)
            QApplication.instance().processEvents()

        filters = self.channel_control.get_filter()
        for i, layer in enumerate(image_info.layers):
            try:
                self._add_layer_util(i, layer, filters)
            except AssertionError:  # noqa: PERF203
                layer.colormap = "gray"
                self._add_layer_util(i, layer, filters)

        self.image_info[image.file_path].filter_info = filters
        self.image_info[image.file_path].layers = image_info.layers
        self.current_image = image.file_path
        mask_layer = self.image_info[image.file_path].mask
        self._add_or_move_layer(mask_layer, -1)
        roi_layer = self.image_info[image.file_path].roi
        self._add_or_move_layer(roi_layer, -1)
        self.viewer.reset_view()
        if self.viewer.layers:
            if hasattr(self.viewer.layers, "selection"):
                self.viewer.layers.selection.clear()
                self.viewer.layers.selection.add(self.viewer.layers[-1])
            else:
                self.viewer.layers[-1].selected = True

        for i, axis in enumerate(image.array_axis_order):
            if axis == "C":
                continue
            self.viewer.dims.set_point(i, image.shape[i] * image.normalized_scaling()[i] // 2)
        if self.image_info[image.file_path].roi is not None:
            self.set_roi()
        if image_info.image.mask is not None:
            self.set_mask()
        self._toggle_mask_chk_visibility()
        self.image_added.emit()

    def add_image(self, image: Optional[Image], replace=False):
        if image is None:
            image = self.settings.image

        if not image.channels:
            raise ValueError("Need non empty image")

        if image.file_path in self.image_info:
            raise ValueError("Image already added")

        self.image_info[image.file_path] = ImageInfo(image, [])

        channels = image.channels
        if self.image_info and not replace:
            channels = max(channels, *(x.image.channels for x in self.image_info.values()))

        self.channel_control.set_channels(channels)
        visibility = self.channel_control.channel_visibility
        limits = self.channel_control.get_limits()
        ranges = image.get_ranges()
        limits = [ranges[i] if x is None else x for i, x in zip(range(image.channels), limits)]
        gamma = self.channel_control.get_gamma()
        colormaps = [self.channel_control.selected_colormaps[i] for i in range(image.channels)]
        parameters = ImageParameters(
            limits, visibility, gamma, colormaps, image.normalized_scaling(), len(self.viewer.layers)
        )

        self._prepare_layers(image, parameters, replace)

        return image

    def _prepare_layers(self, image, parameters, replace):
        worker = prepare_layers(image, parameters, replace)
        worker.returned.connect(self._add_image)
        worker.finished.connect(self._remove_worker)
        self.worker_list.append(worker)
        worker.start()

    def images_bounds(self) -> Tuple[List[int], List[int]]:
        ranges = []
        for image_info in self.image_info.values():
            if not image_info.layers:
                continue
            ranges = [
                (min(a, b), max(c, d), min(e, f))
                for (a, c, e), (b, d, f) in itertools.zip_longest(
                    image_info.layers[0].dims.range, ranges, fillvalue=(np.inf, -np.inf, np.inf)
                )
            ]

        visible = [ranges[i] for i in self.viewer.dims.displayed]
        min_shape, max_shape, _ = zip(*visible)
        size = np.subtract(max_shape, min_shape)
        return size, min_shape

    @staticmethod
    def _shift_layer(layer: Layer, translate_2d):
        translate = [0] * layer.ndim
        translate[-2:] = translate_2d
        layer.translate = translate

    def grid_view(self):
        """Present multiple images in grid view"""
        n_row = np.ceil(np.sqrt(len(self.image_info))).astype(int)
        n_row = max(1, n_row)
        scene_size, _ = self.images_bounds()
        for image_info, pos in zip(self.image_info.values(), itertools.product(range(n_row), repeat=2)):
            translate_2d = np.multiply(scene_size[-2:], pos)
            for layer in image_info.layers:
                self._shift_layer(layer, translate_2d)

            if image_info.mask is not None:
                self._shift_layer(image_info.mask, translate_2d)

            if image_info.roi is not None:
                self._shift_layer(image_info.roi, translate_2d)
        self.viewer.reset_view()

    def change_visibility(self, name: str, index: int):
        for image_info in self.image_info.values():
            if len(image_info.layers) > index:
                image_info.layers[index].visible = self.channel_control.channel_visibility[index]
                if self.channel_control.channel_visibility[index]:
                    image_info.layers[index].colormap = self.channel_control.selected_colormaps[index]
                    limits = self.channel_control.get_limits()[index]
                    limits = image_info.image.get_ranges()[index] if limits is None else limits
                    image_info.layers[index].contrast_limits = limits
                    image_info.layers[index].gamma = self.channel_control.get_gamma()[index]
                    filter_type = self.channel_control.get_filter()[index]
                    if filter_type != image_info.filter_info[index]:
                        image_info.layers[index].data = self.calculate_filter(
                            image_info.image.get_channel(index), filter_type
                        )
                        image_info.filter_info[index] = filter_type

    def reset_image_size(self):
        self.viewer.reset_view()

    def set_theme(self, theme: str):
        self.viewer.theme = theme

    def closeEvent(self, event):
        for worker in self.worker_list:
            worker.quit()
        self.viewer.layers.clear()
        self.viewer_widget.close()
        super().closeEvent(event)

    def get_tool_tip_text(self) -> str:
        image = self.settings.image
        image_info = self.image_info[image.file_path]
        text_list = []
        for el in self.components:
            if data := image_info.roi_info.annotations.get(el, {}):
                try:
                    text_list.append(_print_dict(data))
                except ValueError:  # pragma: no cover
                    logging.warning("Wrong value provided as layer annotation.")
        return " ".join(text_list)

    def event(self, event: QEvent):
        if event.type() == QEvent.Type.ToolTip and self.components and (text := self.get_tool_tip_text()):
            QToolTip.showText(event.globalPos(), text, self)
        return super().event(event)

    def _search_component(self):
        max_components = max(max(image_info.roi_info.bound_info) for image_info in self.image_info.values())
        if self.viewer.dims.ndisplay == 3:
            self._search_type = SearchType.Highlight

        dial = SearchComponentModal(self, self._search_type, self._last_component, max_components)
        if self.viewer.dims.ndisplay == 3:
            dial.zoom_to.setDisabled(True)
        dial.show_right_of_mouse()

    def component_unmark(self, _num):
        for el in self.image_info.values():
            if el.highlight is None:
                continue
            if "timer" in el.highlight.metadata:
                el.highlight.metadata["timer"].stop()
            el.highlight.visible = False

    def _mark_layer(self, num: int, flash: bool, image_info: ImageInfo):
        bound_info = image_info.roi_info.bound_info.get(num, None)
        if bound_info is None:
            return
        # TODO think about marking on bright background
        slices = bound_info.get_slices(1)
        slices[image_info.image.stack_pos] = slice(None)
        component_mark = image_info.roi_info.roi[tuple(slices)] == num
        if self.viewer.dims.ndisplay == 3:
            component_mark = binary_dilation(component_mark)
        shift_base = bound_info.lower - 1
        shift_base[0] += 1  # remove shift on time axis
        translate = image_info.roi.translate + shift_base * image_info.roi.scale
        translate[image_info.image.stack_pos] = 0
        if image_info.highlight is None:
            active_layer = self.viewer.layers.selection.active
            image_info.highlight = self.viewer.add_labels(
                component_mark,
                scale=image_info.roi.scale,
                blending="translucent",
                opacity=0.7,
                **get_highlight_colormap(),
            )
            self.viewer.layers.selection.active = active_layer
        else:
            image_info.highlight.data = component_mark
        image_info.highlight.translate = translate
        image_info.highlight.visible = True
        if flash:
            layer = image_info.highlight

            def flash_fun(layer_=layer):
                opacity = layer_.opacity + 0.1
                if opacity > 1:
                    opacity = 0.1
                layer_.opacity = opacity

            timer = QTimer()
            timer.setInterval(100)
            timer.timeout.connect(flash_fun)
            timer.start()
            layer.metadata["timer"] = timer

    def component_mark(self, num: int, flash: bool = False):
        self.component_unmark(num)
        self._search_type = SearchType.Highlight
        self._last_component = num

        bounding_box = self._bounding_box(num)
        if bounding_box is None:
            return

        for image_info in self.image_info.values():
            self._mark_layer(num, flash, image_info)

        lower_bound, upper_bound = bounding_box
        self._update_point(lower_bound, upper_bound)

        if self.viewer.dims.ndisplay == 2:
            l_bound = lower_bound[-2:][::-1]
            u_bound = upper_bound[-2:][::-1]
            rect = Rect(self.viewer_widget.view.camera.get_state()["rect"])
            if rect.contains(*l_bound) and rect.contains(*u_bound):
                return
            size = u_bound - l_bound
            rect.size = tuple(np.max([rect.size, size * 1.2], axis=0))
            pos = rect.pos
            if rect.left > l_bound[0]:
                pos = l_bound[0], pos[1]
            if rect.right < u_bound[0]:
                pos = pos[0] + u_bound[0] - rect.right, pos[1]
            if rect.bottom > l_bound[1]:
                pos = pos[0], l_bound[1]
            if rect.top < u_bound[1]:
                pos = pos[0], pos[1] + (u_bound[1] - rect.top)
            rect.pos = pos
            self.viewer_widget.view.camera.set_state({"rect": rect})

    def component_zoom(self, num):
        self.component_unmark(num)
        self._search_type = SearchType.Zoom_in
        self._last_component = num

        bounding_box = self._bounding_box(num)
        if bounding_box is None:
            return

        lower_bound, upper_bound = bounding_box
        diff = upper_bound - lower_bound
        frame = diff * (self.settings.get_from_profile(SEARCH_ZOOM_FACTOR_STR, 1.2) - 1)
        if self.viewer.dims.ndisplay == 2:
            rect = Rect(pos=(lower_bound - frame)[-2:][::-1], size=(diff + 2 * frame)[-2:][::-1])
            self.set_state({"camera": {"rect": rect}})
        self._update_point(lower_bound, upper_bound)

    def _update_point(self, lower_bound, upper_bound):
        point = (lower_bound + upper_bound) / 2
        current_point = self.viewer.dims.point
        for i in range(self.viewer.dims.ndim - self.viewer.dims.ndisplay):
            if not (lower_bound[i] <= current_point[i] <= upper_bound[i]):
                self.viewer.dims.set_point(i, point[i])

    @staticmethod
    def _data_to_world(layer: Layer, cords):
        return layer._transforms[1:3].simplified(cords)  # pylint: disable=protected-access

    def _bounding_box(self, num) -> Optional[Tuple[np.ndarray, np.ndarray]]:
        lower_bound_list = []
        upper_bound_list = []
        for image_info in self.image_info.values():
            bound_info = image_info.roi_info.bound_info.get(num, None)
            if bound_info is None:
                continue
            lower_bound_list.append(self._data_to_world(image_info.roi, bound_info.lower))
            upper_bound_list.append(self._data_to_world(image_info.roi, bound_info.upper))

        if not lower_bound_list:
            return None

        lower_bound = np.min(lower_bound_list, axis=0)
        upper_bound = np.min(upper_bound_list, axis=0)
        return lower_bound, upper_bound


class NapariQtViewer(QtViewer):
    def __init__(self, viewer):
        super().__init__(viewer, show_welcome_screen=False)
        self.widget(0).layout().setContentsMargins(0, 5, 0, 2)

    def dragEnterEvent(self, event):  # pylint: disable=no-self-use
        """
        ignore napari reading mechanism
        """
        event.ignore()

    def close(self):
        self.dockConsole.deleteLater()
        self.dockLayerList.deleteLater()
        self.dockLayerControls.deleteLater()
        if hasattr(self, "activityDock"):
            self.activityDock.deleteLater()
        return super().close()

    def closeEvent(self, event):
        self.close()
        super().closeEvent(event)

    def _render(self):
        if _napari_ge_5:
            return self.canvas._scene_canvas.render()
        return self.canvas.render()

    if _napari_ge_5:

        @property
        def view(self):
            return self.canvas.view


class SearchComponentModal(QtPopup):
    def __init__(self, image_view: ImageView, search_type: SearchType, component_num: int, max_components):
        super().__init__(image_view)
        self.image_view = image_view
        self.zoom_to = QEnumComboBox(self, SearchType)
        self.zoom_to.setCurrentEnum(search_type)
        self.zoom_to.currentEnumChanged.connect(self._component_num_changed)
        self.component_selector = QSpinBox()
        self.component_selector.valueChanged.connect(self._component_num_changed)
        self.component_selector.setMaximum(max_components)
        self.component_selector.setValue(component_num)

        layout = QHBoxLayout()
        layout.addWidget(QLabel("Component:"))
        layout.addWidget(self.component_selector)
        layout.addWidget(QLabel("Selection:"))
        layout.addWidget(self.zoom_to)
        self.frame.setLayout(layout)

    def _component_num_changed(self):
        if self.zoom_to.currentEnum() == SearchType.Highlight:
            self.image_view.component_mark(self.component_selector.value(), flash=True)
        else:
            self.image_view.component_zoom(self.component_selector.value())

    def closeEvent(self, event):
        super().closeEvent(event)
        self.image_view.component_unmark(0)


@dataclass
class ImageParameters:
    limits: List[Tuple[float, float]]
    visibility: List[bool]
    gamma: List[float]
    colormaps: List[Colormap]
    scaling: Tuple[Union[float, int]]
    layers: int = 0


def _prepare_layers(image: Image, param: ImageParameters, replace: bool) -> Tuple[ImageInfo, bool]:
    image_layers = []
    for i in range(image.channels):
        lim = list(param.limits[i])
        if lim[1] == lim[0]:
            lim[1] += 1
        blending = "additive" if i != 0 else "translucent"
        data = image.get_channel(i)

        layer = NapariImage(
            data,
            colormap=param.colormaps[i],
            visible=param.visibility[i],
            blending=blending,
            scale=param.scaling,
            contrast_limits=lim,
            gamma=param.gamma[i],
            name=f"channel {i}; {param.layers + i}",
        )
        image_layers.append(layer)
    return ImageInfo(image, image_layers, []), replace


prepare_layers = thread_worker(_prepare_layers)


def _calc_layer_filter(layer: NapariImage, filter_type: NoiseFilterType, radius: float):
    if filter_type == NoiseFilterType.No or radius == 0:
        return None, layer
    return ImageView.calculate_filter(layer.data, parameters=(filter_type, radius)), layer


calc_layer_filter = thread_worker(_calc_layer_filter)


def _print_dict(dkt: MutableMapping, indent="") -> str:
    if not isinstance(dkt, MutableMapping):  # pragma: no cover
        logging.error("%s instead of dict passed to _print_dict", type(dkt))
        return indent + str(dkt)
    res = []
    for k, v in dkt.items():
        if isinstance(v, MutableMapping):
            res.append(f'{indent}{k}:\n{_print_dict(v, f"{indent}  ")}')
        else:
            res.append(f"{indent}{k}: {v}")
    return "\n".join(res)


def _print_list(lst: list, prefix: str) -> str:
    if not lst:
        return ""
    if len(lst) == 1:
        return f"{prefix}: {lst[0]}"
    return f"{prefix}s: {lst}"



from typing import Dict, Iterable, Optional, Tuple

from napari import Viewer
from napari.layers import Image
from napari.utils import Colormap
from qtpy.QtWidgets import QHBoxLayout, QPushButton, QTabWidget

from PartSeg.common_gui.colormap_creator import ChannelPreview, ColormapCreator, ColormapList, save_colormap_in_settings
from PartSeg.plugins.napari_widgets._settings import get_settings
from PartSegCore.custom_name_generate import custom_name_generate


class NapariColormapControl(ChannelPreview):
    def __init__(
        self, viewer: Viewer, colormap: Colormap, accepted: bool, name: str, removable: bool = False, used: bool = False
    ):
        super().__init__(colormap, accepted, name, removable, used)
        self.viewer = viewer
        self.apply_colormap_btn = QPushButton("Apply")
        layout: QHBoxLayout = self.layout()
        layout.removeWidget(self.checked)
        layout.insertWidget(0, self.apply_colormap_btn)

        self.apply_colormap_btn.clicked.connect(self.apply_colormap)
        viewer.layers.selection.events.changed.connect(self.update_preview)
        self.update_preview()

    def update_preview(self, _event=None):
        if len(self.viewer.layers.selection) == 1 and isinstance(next(iter(self.viewer.layers.selection)), Image):
            self.apply_colormap_btn.setEnabled(True)
            self.apply_colormap_btn.setToolTip("Apply colormap to selected layer")
        else:
            self.apply_colormap_btn.setEnabled(False)
            self.apply_colormap_btn.setToolTip("Select one image layer to apply colormap")

    def apply_colormap(self):
        if len(self.viewer.layers.selection) == 1 and isinstance(
            layer := next(iter(self.viewer.layers.selection)), Image
        ):
            layer.colormap = self.colormap


class NapariColormapList(ColormapList):
    def __init__(
        self,
        viewer: Viewer,
        colormap_map: Dict[str, Tuple[Colormap, bool]],
        selected: Optional[Iterable[str]] = None,
        parent=None,
    ):
        super().__init__(colormap_map, selected, parent)
        self.viewer = viewer

    def _create_colormap_preview(self, colormap: Colormap, name: str, removable: bool) -> ChannelPreview:
        return NapariColormapControl(self.viewer, colormap, False, name, removable)


class ImageColormap(QTabWidget):
    def __init__(self, viewer: Viewer, parent=None):
        super().__init__(parent)
        self.viewer = viewer
        settings = get_settings()
        self.settings = settings
        self.colormap_list = NapariColormapList(viewer, settings.colormap_dict, parent=self)
        self.colormap_creator = ColormapCreator(self)
        self.addTab(self.colormap_list, "List")
        self.addTab(self.colormap_creator, "Creator")

        self.colormap_creator.colormap_selected.connect(self.handle_new_colormap)

        self.colormap_list.edit_signal.connect(self.colormap_creator.set_colormap)
        self.colormap_list.edit_signal.connect(self._set_colormap_editor)

    def handle_new_colormap(self, colormap):
        rand_name = custom_name_generate(set(), self.settings.colormap_dict)
        save_colormap_in_settings(self.settings, colormap, rand_name)
        self.settings.dump()

    def _set_colormap_editor(self):
        self.setCurrentWidget(self.colormap_creator)



import os
import subprocess  # nosec
import sys

from qtpy.QtCore import Qt
from qtpy.QtWidgets import QDialog, QGridLayout, QLabel, QPushButton


class DirectoryDialog(QDialog):
    def __init__(self, path_to_directory, additional_text=""):
        super().__init__()
        self.setWindowTitle("Path dialog")
        self.path_to_directory = path_to_directory
        text_label = QLabel(path_to_directory)
        text_label.setWordWrap(True)
        text_label.setTextInteractionFlags(Qt.TextInteractionFlag.TextSelectableByMouse)
        open_btn = QPushButton("Open directory")
        open_btn.clicked.connect(self.open_folder)
        close_btn = QPushButton("Close")
        close_btn.clicked.connect(self.accept)
        layout = QGridLayout()
        if additional_text:
            layout.addWidget(QLabel(f"{additional_text}<br><br>Path:"), 0, 0, 1, 2)
        else:
            layout.addWidget(QLabel("Path:"), 0, 0, 1, 2)
        layout.addWidget(text_label, 1, 0, 1, 2)
        layout.addWidget(close_btn, 2, 0)
        layout.addWidget(open_btn, 2, 1)
        self.setLayout(layout)

    def open_folder(self):
        if sys.platform in ["linux", "linux2"]:
            subprocess.Popen(["xdg-open", self.path_to_directory])  # nosec
        elif sys.platform == "darwin":
            subprocess.Popen(["open", self.path_to_directory])  # nosec
        elif sys.platform == "win32":
            os.startfile(self.path_to_directory)  # nosec
        self.accept()



import os
import re
from functools import partial
from pathlib import Path
from queue import Queue
from typing import List, NamedTuple, Optional, Tuple, Union, cast

from pydantic import BaseModel
from qtpy.QtCore import QThread, Signal

from PartSeg._roi_mask.stack_settings import StackSettings, get_mask
from PartSegCore.algorithm_describe_base import ROIExtractionProfile
from PartSegCore.mask.algorithm_description import MaskAlgorithmSelection
from PartSegCore.mask.io_functions import LoadROIImage, LoadStackImage, MaskProjectTuple, SaveROI
from PartSegCore.segmentation import StackAlgorithm
from PartSegCore.segmentation.algorithm_base import ROIExtractionAlgorithm


class BatchTask(NamedTuple):
    data: Union[str, MaskProjectTuple]
    parameters: ROIExtractionProfile
    save_prefix: Optional[Tuple[Union[str, Path], Union[dict, BaseModel]]]


class BatchProceed(QThread):
    error_signal = Signal(str)
    progress_signal = Signal(str, int, str, int)
    range_signal = Signal(int, int)
    execution_done = Signal()
    multiple_result = Signal(MaskProjectTuple)
    algorithm: ROIExtractionAlgorithm

    def __init__(self):
        super().__init__()
        self.queue = Queue()
        self.algorithm = Optional[None]
        self.parameters = None
        self.file_list = []
        self.index = 0
        self.result_dir = ""
        self.save_parameters = {}

    def add_task(self, task: Union[BatchTask, List[BatchTask]]):
        if isinstance(task, list):
            for el in task:
                self.queue.put(el)
        else:
            self.queue.put(task)

    def progress_info(self, name: str, text: str, num: int):
        self.progress_signal.emit(text, num, name, self.index)

    def run_calculation(self):  # noqa: PLR0912  # FIXME
        while not self.queue.empty():
            task: BatchTask = self.queue.get()
            if isinstance(task.data, str):
                file_path = task.data
                if os.path.splitext(task.data)[1] in LoadROIImage.get_extensions():
                    project_tuple = LoadROIImage.load([task.data])
                else:
                    project_tuple = LoadStackImage.load([task.data])
            elif isinstance(task.data, MaskProjectTuple):
                project_tuple: MaskProjectTuple = task.data
                file_path = project_tuple.image.file_path
            else:
                continue
            try:
                name = os.path.basename(file_path)
                blank = get_mask(project_tuple.roi_info.roi, project_tuple.mask, project_tuple.selected_components)
                algorithm = cast(StackAlgorithm, MaskAlgorithmSelection[task.parameters.algorithm]())
                algorithm.set_image(project_tuple.image)
                algorithm.set_mask(blank)
                algorithm.set_parameters(task.parameters.values)
                if isinstance(task.save_prefix, tuple):
                    self.range_signal.emit(0, algorithm.get_steps_num() + 1)
                else:
                    self.range_signal.emit(0, algorithm.get_steps_num())
                # noinspection PyTypeChecker
                segmentation = algorithm.calculation_run(partial(self.progress_info, name))
                state2 = StackSettings.transform_state(
                    project_tuple,
                    segmentation.roi_info,
                    (
                        {i: segmentation.parameters for i in segmentation.roi_info.bound_info}
                        if segmentation.roi_info is not None
                        else {}
                    ),
                    [],
                )
                if isinstance(task.save_prefix, tuple):
                    self.progress_info(name, "saving", algorithm.get_steps_num())
                    name = f"{os.path.splitext(os.path.basename(file_path))[0]}.seg"
                    re_end = re.compile(r"(.*_version)(\d+)\.seg$")
                    os.makedirs(task.save_prefix[0], exist_ok=True)
                    while os.path.exists(os.path.join(task.save_prefix[0], name)):
                        if match := re_end.match(name):
                            num = int(match[2]) + 1
                            name = match[1] + str(num) + ".seg"
                        else:
                            name = f"{os.path.splitext(os.path.basename(file_path))[0]}_version1.seg"
                    SaveROI.save(os.path.join(task.save_prefix[0], name), state2, parameters=task.save_prefix[1])
                else:
                    self.multiple_result.emit(state2)
            except Exception as e:  # pylint: disable=broad-except
                self.error_signal.emit(f"Exception occurred during proceed {file_path}. Exception info {e}")
            self.index += 1
        self.index = 0
        self.execution_done.emit()

    def run(self):
        self.index = 0
        self.run_calculation()



from enum import Enum
from typing import Callable, List, Optional, Tuple

import numpy as np
from pydantic import Field

from PartSegCore.image_transforming.transform_base import TransformBase
from PartSegCore.roi_info import ROIInfo
from PartSegCore.utils import BaseModel
from PartSegImage import Image


class ProjectionType(Enum):
    MAX = "max"
    MIN = "min"
    MEAN = "mean"
    SUM = "sum"


class ImageProjectionParams(BaseModel):
    projection_type: ProjectionType = Field(ProjectionType.MAX, suffix="Mask and ROI projection will always use max")
    keep_mask: bool = False
    keep_roi: bool = False


def _calc_target_shape(image: Image):
    new_shape = list(image.shape)
    new_shape[image.array_axis_order.index("Z")] = 1
    return tuple(new_shape)


class ImageProjection(TransformBase):
    __argument_class__ = ImageProjectionParams

    @classmethod
    def transform(
        cls,
        image: Image,
        roi_info: ROIInfo,
        arguments: ImageProjectionParams,  # type: ignore[override]
        callback_function: Optional[Callable[[str, int], None]] = None,
    ) -> Tuple[Image, Optional[ROIInfo]]:
        project_operator = getattr(np, arguments.projection_type.value)
        axis = image.array_axis_order.index("Z")
        target_shape = _calc_target_shape(image)
        spacing = list(image.spacing)
        spacing.pop(axis - 1 if image.time_pos < image.stack_pos else axis)
        new_channels = [
            project_operator(image.get_channel(i), axis=axis).reshape(target_shape) for i in range(image.channels)
        ]
        new_mask = None
        if arguments.keep_mask and image.mask is not None:
            new_mask = np.max(image.mask, axis=axis).reshape(target_shape)

        roi = None
        if arguments.keep_roi and roi_info.roi is not None:
            roi = ROIInfo(np.max(image.fit_array_to_image(roi_info.roi), axis=axis).reshape(target_shape))
        return (
            image.__class__(
                data=new_channels,
                image_spacing=tuple(spacing),
                channel_names=image.channel_names,
                mask=new_mask,
                axes_order=image.axis_order,
            ),
            roi,
        )

    @classmethod
    def get_fields_per_dimension(cls, component_list: List[str]):
        return cls.__argument_class__

    @classmethod
    def calculate_initial(cls, image: Image):
        return cls.get_default_values()

    @classmethod
    def get_name(cls) -> str:
        return "Image Projection"



import dataclasses
import os
import sys
from pathlib import Path
from typing import List, Optional, Type, Union

from napari import __version__
from packaging.version import parse as parse_version
from qtpy.QtCore import Qt, Signal
from qtpy.QtGui import QCloseEvent, QDragEnterEvent, QDropEvent, QShowEvent
from qtpy.QtWidgets import QAction, QApplication, QDockWidget, QMainWindow, QMenu, QMessageBox, QWidget
from vispy.color import colormap

from PartSeg.common_backend.base_settings import (
    FILE_HISTORY,
    BaseSettings,
    SwapTimeStackException,
    TimeAndStackException,
)
from PartSeg.common_backend.except_hook import show_warning
from PartSeg.common_backend.load_backup import import_config
from PartSeg.common_gui.about_dialog import AboutDialog
from PartSeg.common_gui.custom_save_dialog import PSaveDialog
from PartSeg.common_gui.error_report import DataImportErrorDialog
from PartSeg.common_gui.exception_hooks import load_data_exception_hook
from PartSeg.common_gui.image_adjustment import ImageAdjustmentDialog
from PartSeg.common_gui.napari_image_view import ImageView
from PartSeg.common_gui.napari_viewer_wrap import Viewer
from PartSeg.common_gui.qt_console import QtConsole
from PartSeg.common_gui.show_directory_dialog import DirectoryDialog
from PartSeg.common_gui.waiting_dialog import ExecuteFunctionDialog
from PartSegCore.algorithm_describe_base import Register
from PartSegCore.io_utils import LoadBase, SaveScreenshot
from PartSegCore.project_info import ProjectInfoBase
from PartSegImage import Image

OPEN_FILE = "io.open_file"
OPEN_DIRECTORY = "io.open_directory"
OPEN_FILE_FILTER = "io.open_filter"

_EXIT = object()

NAPARI_LE_4_16 = parse_version(__version__) <= parse_version("0.4.16")


class BaseMainMenu(QWidget):
    def __init__(self, settings: BaseSettings, main_window):
        super().__init__()
        self.settings = settings
        self.main_window = main_window

    def _set_data_list(self, data_list):
        if len(data_list) == 0:
            QMessageBox.warning(self, "Empty list", "List of files to load is empty")
            return _EXIT
        if hasattr(self.main_window, "multiple_files"):
            self.main_window.multiple_files.add_states(data_list)
            self.main_window.multiple_files.setVisible(True)
            self.settings.set("multiple_files", True)
        return data_list[0]

    def _set_project_info_base(self, project_info_base: ProjectInfoBase):
        if project_info_base.errors:  # pragma: no cover
            resp = QMessageBox.question(
                self,
                "Load problem",
                f"During load data "
                f"some problems occur: {project_info_base.errors}."
                "Do you would like to try load it anyway?",
                QMessageBox.Yes | QMessageBox.No,
            )
            if resp == QMessageBox.No:
                return _EXIT
        try:
            image = self.settings.verify_image(project_info_base.image, False)
        except SwapTimeStackException:
            res = QMessageBox.question(
                self,
                "Not supported",
                "Time data are currently not supported. Maybe You would like to treat time as z-stack",
                QMessageBox.Yes | QMessageBox.No,
                QMessageBox.No,
            )

            if res == QMessageBox.Yes:
                image = project_info_base.image.swap_time_and_stack()
            else:
                return _EXIT
        except TimeAndStackException:
            QMessageBox.warning(self, "image error", "Do not support time and stack image")
            return _EXIT
        if not image:
            return _EXIT
        if isinstance(image, Image):
            # noinspection PyProtectedMember
            project_info_base = dataclasses.replace(project_info_base, image=image)
        return project_info_base

    def set_data(self, data):
        if isinstance(data, list):
            data = self._set_data_list(data)
        if isinstance(data, ProjectInfoBase):
            data = self._set_project_info_base(data)
        if data is _EXIT:
            return
        if data is None:
            QMessageBox().warning(self, "Data loading failure", "Error during data loading", QMessageBox.Ok)
            return
        self.settings.set_project_info(data)
        if data.file_path:
            self.settings.add_path_history(os.path.dirname(data.file_path))


class BaseMainWindow(QMainWindow):
    """
    Base for main windows of subprograms

    :ivar settings: store state of application. initial value is obtained from :py:attr:`.settings_class`
    :ivar files_num: maximal number of files accepted by drag and rop event
    :param config_folder: path to directory in which application save state. If `settings` parameter is note
        then settings object is created with passing this path to :py:attr:`.settings_class`.
        If this parameter and `settings`
        are None then constructor fail with :py:exc:`ValueError`.
    :param title: Window default title
    :param settings: object to store application state
    :param signal_fun: function which need to be called when window shown.
    """

    show_signal = Signal()
    """Signal emitted when window has shown. Used to hide Launcher."""

    @classmethod
    def get_setting_class(cls) -> Type[BaseSettings]:
        """Get constructor for :py:attr:`settings`"""
        return BaseSettings

    def __init__(
        self,
        config_folder: Union[str, Path, None] = None,
        title="PartSeg",
        settings: Optional[BaseSettings] = None,
        load_dict: Optional[Register] = None,
        signal_fun=None,
    ):
        if settings is None:
            if config_folder is None:
                raise ValueError("wrong config folder")
            if not os.path.exists(config_folder):
                import_config()
            settings: BaseSettings = self.get_setting_class()(config_folder)
            if errors := settings.load():  # pragma: no cover
                DataImportErrorDialog(
                    errors,
                    text="During load saved state some of data could not be load properly\n"
                    "The files has prepared backup copies in "
                    " state directory (Help > State directory)",
                ).exec_()

        super().__init__()
        if signal_fun is not None:
            self.show_signal.connect(signal_fun)
        self.settings = settings
        self._load_dict = load_dict
        self.viewer_list: List[Viewer] = []
        self.files_num = 1
        self.setAcceptDrops(True)
        self.setWindowTitle(title)
        self.title_base = title
        app = QApplication.instance()
        if app is not None and "PYTEST_CURRENT_TEST" not in os.environ:  # pragma: no cover
            # FIXME the PYTEST_CURRENT_TEST is used to prevent pytest session fail on CI.
            app.setStyleSheet(settings.get_style_sheet())
        self.settings.theme_changed.connect(self.change_theme)
        self.channel_info = ""
        self.multiple_files = None
        self.settings.request_load_files.connect(self.read_drop)
        self.recent_file_menu = QMenu("Open recent")
        self._refresh_recent()
        self.settings.connect_(FILE_HISTORY, self._refresh_recent)
        self.settings.napari_settings.appearance.events.theme.connect(self.change_theme)
        self.settings.set_parent(self)
        self.console = None
        self.console_dock = QDockWidget("console", self)
        self.console_dock.setAllowedAreas(Qt.LeftDockWidgetArea | Qt.BottomDockWidgetArea)
        self.console_dock.hide()
        self.addDockWidget(Qt.BottomDockWidgetArea, self.console_dock)

        self._scale_bar_warning = True  # remove after drop napari 0.4.16

    def _toggle_console(self):
        if self.console is None:
            self.console = QtConsole(self)
            self.console_dock.setWidget(self.console)
        self.console_dock.setVisible(not self.console_dock.isVisible())

    def _refresh_recent(self):
        self.recent_file_menu.clear()
        for name_list, method in self.settings.get_last_files():
            action = self.recent_file_menu.addAction(f"{name_list[0]}, {method}")
            action.setData((name_list, method))
            action.triggered.connect(self._load_recent)

    def _load_recent(self):
        sender: QAction = self.sender()
        data = sender.data()
        try:
            method: LoadBase = self._load_dict[data[1]]
            dial = ExecuteFunctionDialog(method.load, [data[0]], exception_hook=load_data_exception_hook)
            if dial.exec_():
                result = dial.get_result()
                self.main_menu.set_data(result)
                self.settings.add_last_files(data[0], method.get_name())
                self.settings.set(OPEN_DIRECTORY, os.path.dirname(data[0][0]))
                self.settings.set(OPEN_FILE, data[0][0])
                self.settings.set(OPEN_FILE_FILTER, data[1])
        except KeyError:  # pragma: no cover
            self.read_drop(data[0])

    def toggle_multiple_files(self):
        self.settings.set("multiple_files_widget", not self.settings.get("multiple_files_widget", False))

    def get_colormaps(self) -> List[Optional[colormap.Colormap]]:
        channel_num = self.settings.image.channels
        if not self.channel_info:
            return [None for _ in range(channel_num)]
        colormaps_name = [self.settings.get_channel_colormap_name(self.channel_info, i) for i in range(channel_num)]
        return [self.settings.colormap_dict[name][0] for name in colormaps_name]

    def napari_viewer_show(self):
        viewer = Viewer(title="Additional output", settings=self.settings, partseg_viewer_name=self.channel_info)
        viewer.theme = self.settings.theme_name
        viewer.create_initial_layers(image=True, roi=True, additional_layers=False, points=True)
        self.viewer_list.append(viewer)
        viewer.window.qt_viewer.destroyed.connect(lambda _x: self.close_viewer(viewer))

    def additional_layers_show(self, with_channels=False):
        if not self.settings.additional_layers:
            QMessageBox().information(self, "No data", "Last executed algorithm does not provide additional data")
            return
        viewer = Viewer(title="Additional output", settings=self.settings, partseg_viewer_name=self.channel_info)
        viewer.theme = self.settings.theme_name
        viewer.create_initial_layers(image=with_channels, roi=False, additional_layers=True, points=False)
        self.viewer_list.append(viewer)
        viewer.window.qt_viewer.destroyed.connect(lambda _x: self.close_viewer(viewer))

    def close_viewer(self, obj):
        for i, el in enumerate(self.viewer_list):
            if el == obj:
                self.viewer_list.pop(i)
                break

    # @ensure_main_thread
    def change_theme(self, event):
        style_sheet = self.settings.style_sheet
        app = QApplication.instance()
        if app is not None and "PYTEST_CURRENT_TEST" not in os.environ:  # pragma: no cover
            app.setStyleSheet(style_sheet)
        self.setStyleSheet(style_sheet)

    def showEvent(self, a0: QShowEvent):
        self.show_signal.emit()

    def dragEnterEvent(self, event: QDragEnterEvent):  # pylint: disable=no-self-use
        if event.mimeData().hasUrls():
            event.acceptProposedAction()

    def read_drop(self, paths: List[str]):
        """Function to process loading files by drag and drop."""
        self._read_drop(paths, self._load_dict)

    def _read_drop(self, paths, load_dict):
        ext_set = {os.path.splitext(x)[1].lower() for x in paths}

        def exception_hook(exception):  # pragma: no cover
            additional_info = "files: " + ", ".join(paths)

            if isinstance(exception, OSError):
                # if happens on macos then add information about requirements to check permissions to file
                if sys.platform == "darwin":
                    additional_info += (
                        "In latest macos release you may need to check if you gave PartSeg (or terminal)"
                        "Permission to access files. You can do it in System Preferences -> Security & Privacy"
                    )
                show_warning(
                    "IO Error",
                    "Disc operation error: " + ", ".join(str(x) for x in exception.args) + additional_info,
                    exception=exception,
                )
            else:
                raise exception

        for load_class in load_dict.values():
            if load_class.partial() or load_class.number_of_files() != len(paths):
                continue
            if ext_set.issubset(load_class.get_extensions()):
                dial = ExecuteFunctionDialog(load_class.load, [paths], exception_hook=exception_hook)
                if dial.exec_():
                    result = dial.get_result()
                    self.main_menu.set_data(result)
                    self.settings.add_last_files(paths, load_class.get_name())
                return
        QMessageBox.information(self, "No method", "No methods for load files: " + ",".join(paths))

    def dropEvent(self, event: QDropEvent):
        """
        Support for load files by drag and drop.
        At beginning it check number of files and if it greater than :py:attr:`.files_num` it refuse loading. Otherwise
        it call :py:meth:`.read_drop` method and this method should be overwritten in sub classes
        """
        if not all(x.isLocalFile() for x in event.mimeData().urls()):
            QMessageBox().warning(self, "Load error", "Not all files are locally. Cannot load data.", QMessageBox.Ok)
        paths = [x.toLocalFile() for x in event.mimeData().urls()]
        if self.files_num != -1 and len(paths) > self.files_num:
            QMessageBox.information(self, "To many files", "currently support only drag and drop one file")
            return
        self.read_drop(paths)

    def show_settings_directory(self):
        DirectoryDialog(
            self.settings.json_folder_path, "Path to place where PartSeg store the data between runs"
        ).exec_()

    @staticmethod
    def show_about_dialog():
        """Show about dialog."""
        AboutDialog().exec_()

    @staticmethod
    def get_project_info(file_path, image, roi_info=None):
        raise NotImplementedError

    def image_adjust_exec(self):
        dial = ImageAdjustmentDialog(self.settings.image)
        if dial.exec_():
            algorithm = dial.result_val.algorithm
            dial2 = ExecuteFunctionDialog(
                algorithm.transform,
                [],
                {"image": self.settings.image, "arguments": dial.result_val.values, "roi_info": self.settings.roi_info},
            )
            if dial2.exec_():
                image, roi_info = dial2.get_result()
                self.settings.set_project_info(self.get_project_info(image.file_path, image, roi_info))

    def closeEvent(self, event: QCloseEvent):
        for el in self.viewer_list:
            el.close()
            del el
        self.settings.napari_settings.appearance.events.theme.disconnect(self.change_theme)
        self.settings.dump()
        super().closeEvent(event)

    def screenshot(self, viewer: ImageView):
        def _screenshot():
            data = viewer.viewer_widget.screenshot()
            dial = PSaveDialog(
                SaveScreenshot,
                settings=self.settings,
                system_widget=False,
                path="io.save_screenshot",
                file_mode=PSaveDialog.AnyFile,
            )

            if not dial.exec_():
                return
            res = dial.get_result()
            res.save_class.save(res.save_destination, data, res.parameters)

        return _screenshot

    def image_read(self):
        if self.settings.image_path is None:
            self.setWindowTitle(f"{self.title_base}")
            return
        folder_name, file_name = os.path.split(self.settings.image_path)
        self.setWindowTitle(f"{self.title_base}: {os.path.join(os.path.basename(folder_name), file_name)}")
        self.statusBar().showMessage(self.settings.image_path)

    def deleteLater(self) -> None:
        self.settings.napari_settings.appearance.events.theme.disconnect(self.change_theme)
        super().deleteLater()

    def _toggle_scale_bar(self):
        """Remove after drop napari 0.4.16"""
        if NAPARI_LE_4_16 and self._scale_bar_warning and self.settings.theme_name == "light":  # pragma: no cover
            QMessageBox.warning(
                self, "Not supported", "Scale bar is not supported for light theme and napari below 0.4.17"
            )
            self._scale_bar_warning = False



import json
import urllib.request
from datetime import date, timedelta
from io import StringIO
from unittest.mock import MagicMock

import packaging.version
import pytest
from qtpy.QtWidgets import QMessageBox

from PartSeg import state_store
from PartSeg._launcher import check_version
from PartSeg._launcher.check_version import IGNORE_FILE


@pytest.mark.enablethread()
@pytest.mark.parametrize("thread", [False, True])
@pytest.mark.parametrize("package_name", ["PartSeg", "sample_name"])
def test_fetching(thread, package_name, monkeypatch, qtbot):
    def urlopen_mock(url):
        assert f"https://pypi.org/pypi/{package_name}/json" == url
        return StringIO(
            json.dumps(
                {"info": {"version": "0.10.0", "home_page": f"https://4dnucleome.cent.uw.edu.pl/{package_name}/"}}
            )
        )

    monkeypatch.setattr(urllib.request, "urlopen", urlopen_mock)
    assert packaging.version.parse("0.10.0") < packaging.version.parse("0.11.0")
    chk_thr = check_version.CheckVersionThread(package_name, base_version="0.11.0")
    chk_thr.release = "0.10.0"
    if thread:
        with qtbot.wait_signal(chk_thr.finished):
            chk_thr.start()
    else:
        chk_thr.run()
    assert chk_thr.release == "0.10.0"
    assert chk_thr.url == f"https://4dnucleome.cent.uw.edu.pl/{package_name}/"
    chk_thr.deleteLater()


@pytest.mark.parametrize("frozen", [True, False])
def test_show_window_dialog(monkeypatch, frozen, qtbot):
    values = ["", ""]

    class MockMessageBox:
        StandardButton = QMessageBox.StandardButton

        Icon = QMessageBox.Icon

        def __init__(self, _type, title, message, _buttons):
            values[0] = title
            values[1] = message

        @staticmethod
        def exec_():
            return check_version.QMessageBox.StandardButton.Ok

    chk_thr = check_version.CheckVersionThread(base_version="0.10.0")
    chk_thr.release = "0.11.0"
    monkeypatch.setattr(check_version.sys, "frozen", frozen, raising=False)
    monkeypatch.setattr(check_version, "QMessageBox", MockMessageBox)
    chk_thr.show_version_info()
    assert values[0] == "New release"
    if frozen:
        assert "You can download next release form" in values[1]
    else:
        assert "You can update it from pypi" in values[1]


def test_no_update(monkeypatch, qtbot):
    monkeypatch.setattr("PartSeg.state_store.save_folder", False)

    def urlopen_mock(url):
        raise RuntimeError

    monkeypatch.setattr(urllib.request, "urlopen", urlopen_mock)
    chk_thr = check_version.CheckVersionThread()
    chk_thr.run()


def test_error_report(monkeypatch, qtbot):
    sentry_val = [False]

    def urlopen_mock(url):
        raise RuntimeError

    def sentry_mock(e):
        sentry_val[0] = True

    monkeypatch.setattr(urllib.request, "urlopen", urlopen_mock)
    monkeypatch.setattr(check_version.sentry_sdk, "capture_exception", sentry_mock)
    chk_thr = check_version.CheckVersionThread()
    chk_thr.run()

    assert sentry_val[0] is True


def test_ignore_file_exists(monkeypatch, qtbot, tmp_path):
    with (tmp_path / IGNORE_FILE).open("w") as f_p:
        f_p.write(date.today().isoformat())

    monkeypatch.setattr(state_store, "save_folder", tmp_path)

    def urlopen_mock(url):
        raise RuntimeError

    monkeypatch.setattr(urllib.request, "urlopen", urlopen_mock)

    chk_thr = check_version.CheckVersionThread()
    chk_thr.run()


def test_ignore_file_exists_old_date(monkeypatch, qtbot, tmp_path):
    with (tmp_path / IGNORE_FILE).open("w") as f_p:
        f_p.write((date.today() - timedelta(days=60)).isoformat())

    monkeypatch.setattr(state_store, "save_folder", tmp_path)

    def urlopen_mock(_url):
        return StringIO("")

    monkeypatch.setattr(urllib.request, "urlopen", urlopen_mock)

    chk_thr = check_version.CheckVersionThread()
    chk_thr.run()
    assert not (tmp_path / IGNORE_FILE).exists()


def test_create_ignore(qtbot, tmp_path, monkeypatch):
    monkeypatch.setattr(state_store, "save_folder", tmp_path)

    chk_thr = check_version.CheckVersionThread(base_version="0.10.0")
    chk_thr.release = "0.11.0"
    monkeypatch.setattr(check_version.QMessageBox, "exec_", MagicMock(return_value=check_version.QMessageBox.Ignore))
    chk_thr.show_version_info()
    assert (tmp_path / IGNORE_FILE).exists()
    with (tmp_path / IGNORE_FILE).open() as f_p:
        assert f_p.read() == date.today().isoformat()
